//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                              6545 CRTC module                              *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Emulate the 6545 CRT.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.1.0 - 27 October 2007, S.J.Kay
// - Reversed the full and half intensity colours in the col_table as these
//   were in the wrong order and re-adjusted the intensity levels to suit.
// - Added another intensity level to generate dark grey (was black).
// - Fiddled colour values (in crtc.h) to get better matching colours.
// - 6545 light pen keys emulation option added for 256tc model.
// - vblank status now in crtc_vblank() so that it can be accessed by the pio.c
//   module when used for port B bit 7.
// - Implement the modelx information structure.
//
// v2.0.0 - 14 October 2007, S.J.Kay
// - keyb_checkall() changed to keystd_checkall() and keyb_handler() changed
//   to keystd_handler().  These functions only called if not emulataing the
//   256TC model.
//
// v1.4.0 - 29 September 2007, S.J.Kay
// - Changes to Y scaling to achieve better aspect ratio.
// - Coding improvements to opt_monitor and opt_colour.  Some code tidy up.
// - Removed the non aplpha+ source build option.
//
// v1.0.0 - 14 August 2007, S.J.Kay
// - Added crtc_clock function to set xtal frequency and methods to be used for
//   the VBLANK status.  Now uses Z80 cycles or host timer to gererate VBLANK.
//   The host timer method is needed to keep VBLANK at 50 Hz in turbo mode.
// - Modifed the CRTC vertical blanking status method used.  The new method
//   now gives the correct delays needed for keyboard encoding and some games.
// - Implemented cursor modes (R10 bits 5,6), includes no cursor, steady cursor
//   and blinking cursor.
// - Writes to the CRTC registers (R1, R6, R9) will now cause the display to
//   be resized automatically as required.
// - Implemented the alpha+ inverse and flashing attribute bits.
// - Added standard and standard+colour model options.
// - Implemeted alpha+ (Premium) graphics in functions crtc_redraw and
//   crtc_redraw_char.
// - Added amber, green, black and white monitor options, default is colour.
//
// v0.0.0 - 21 June 2007, S.J.Kay
// - Start with "nanowasp" source distribution version 0.22. An emulator for
//   the microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <SDL.h>

#if MINGW
#else
#include <sys/time.h>
#endif

#include "macros.h"
#include "crtc.h"
#include "keystd.h"
#include "vdu.h"
#include "gui.h"
#include "ubee512.h"
#include "mz80.h"

//==============================================================================
// structures and variables
//==============================================================================
static int crtc_y_scale_conf=1;
int crtc_y_scale=1;

int disp_start;
int crtcresized;

int hdisp=80;
int vdisp=25;
int scans_per_row=11;

int cur_blink;
int cur_mode;
int cur_pos;

static int vblank_divval = 33750L;
static int vblank_method = 0;

static int htot, vtot, vtot_adj;
static int cur_start, cur_end;
static int lpen_valid, lpen;
static int reg;

static int mem_addr;

static SDL_Surface *vdu;
static SDL_Rect rect_vdu = { 0, 0, VDUWIDTH, VDUHEIGHT };
static SDL_Surface *screen;
static SDL_Rect rect_screen = { VDUPOSX, VDUPOSY, VDUWIDTH, VDUHEIGHT };
static SDL_PixelFormat *spf;

static int redraw;

static Uint32 bgcolour;
static Uint32 fgcolour;

static int monochrome;

extern int model;
extern model_t modelx;

extern int monitor;
extern int flashvideo;

extern char charrom[];

extern char screen_ram_ap[];
extern char colour_ram_ap[];
extern char attrib_ram_ap[];
extern char pcg_ram_ap[];

extern int attribram;
extern int extendram;
extern int colourram;

extern int latchrom;

extern uint64_t z80_cycles;

extern FILE *debuglog;
extern int debuglvl;

const UINT8 col_table[16][3] =
{
//   R     G     B
 {   0,    0,    0},		// 00 - black
 {C_HI,    0,    0}, 		// 01 - red
 {   0, C_HI,    0}, 		// 02 - green
 { 225,  130,    0}, 		// 03 - yellow (brown ?)
 {   0,    0, C_HI},		// 04 - blue
 {C_HI,    0, C_HI},		// 05 - magenta
 {   0, C_HI, C_HI},		// 06 - cyan 
 {C_HI, C_HI, C_HI},		// 07 - light grey

 {C_LI, C_LI, C_LI},		// 08 - dark grey
 {C_FI,    0,    0}, 		// 09 - light red
 {   0, C_FI,    0}, 		// 10 - light green
 {C_FI, C_FI,    0}, 		// 11 - light yellow
 {   0,    0, C_FI},		// 12 - light blue
 {C_FI,    0, C_FI},		// 13 - light magenta
 {   0, C_FI, C_FI},		// 14 - light cyan
 {C_FI, C_FI, C_FI}		// 15 - white
}; 

//==============================================================================
// CRTC Create RGB Surface
//
//   pass: int w		width
//         int h		height
// return: int			0 if success, -1 if error
//==============================================================================
int crtc_CreateRGBSurface (int w, int h)
{
 h = h * crtc_y_scale;

 vdu = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, 16, 0, 0, 0, 0);
 screen = SDL_GetVideoSurface();
 spf = screen->format;

 if (vdu == NULL)
    {
     fprintf(stderr, "crtc_init: SDL_CreateRGBSurface failed - %s\n", SDL_GetError());
     return -1;
    }
    
 return 0;   
}

//==============================================================================
// CRTC set the rect_vdu and rect_screen (SDL_Rect) values using the current
// CRTC values.
//
//   pass: void
// return: void
//==============================================================================
void crtc_SetRectValues (void)
{
 rect_vdu.x = VDUPOSX;	// x position of the upper-left corner of the rectangle
 rect_vdu.y = VDUPOSY;  // y position of the upper-left corner of the rectangle
 rect_vdu.w = hdisp * 8;  		// The width of the rectangle

 rect_vdu.h = vdisp * scans_per_row * crtc_y_scale; 	// The height of the rectangle
 rect_screen.x = VDUPOSX; // x position of the upper-left corner of the rectangle
 rect_screen.y = VDUPOSY; // y position of the upper-left corner of the rectangle
 rect_screen.w = hdisp * 8;  		// The width of the rectangle

 rect_screen.h = vdisp * scans_per_row * crtc_y_scale; // The height of the rectangle
}
    
//==============================================================================
// CRTC set monitor
//
//   pass: void
// return: void
//==============================================================================
void crtc_setmonitor (void)
{
// Default to the "classic" green monitor values for display colours
 fgcolour = SDL_MapRGB(spf, MONGR_FGR, MONGR_FGG, MONGR_FGB);
 bgcolour = SDL_MapRGB(spf, MONGR_BGR, MONGR_BGG, MONGR_BGB);

 monochrome = 0;

// if standard model with no colour hardware
 if ((modelx.standard) && (modelx.colour == 0))
    monochrome = 1;

// set the colour scheme according to the moinitor type
 switch (monitor)
    {
     case 0 : // colour (default)
              break;
     case 1 : // amber on black
              fgcolour = SDL_MapRGB(spf, MONAM_FGR, MONAM_FGG, MONAM_FGB);
              bgcolour = SDL_MapRGB(spf, MONAM_BGR, MONAM_BGG, MONAM_BGB);
              monochrome = 1;
              break;
     case 2 : // green on black
              fgcolour = SDL_MapRGB(spf, MONGR_FGR, MONGR_FGG, MONGR_FGB);
              bgcolour = SDL_MapRGB(spf, MONGR_BGR, MONGR_BGG, MONGR_BGB);
              monochrome = 1;
              break;
     case 3 : // black on white
              fgcolour = SDL_MapRGB(spf, MONBW_FGR, MONBW_FGG, MONBW_FGB);
              bgcolour = SDL_MapRGB(spf, MONBW_BGR, MONBW_BGG, MONBW_BGB);
              monochrome = 1;
              break;
     case 4 : // white on black              
              fgcolour = SDL_MapRGB(spf, MONWB_FGR, MONWB_FGG, MONWB_FGB);
              bgcolour = SDL_MapRGB(spf, MONWB_BGR, MONWB_BGG, MONWB_BGB);
              monochrome = 1;
              break;
    }
 
 SDL_FillRect(vdu, &rect_vdu, bgcolour);
}

//==============================================================================
// CRTC Change Video
//
//   pass: void
// return: int			0 if success, -1 if error
//==============================================================================
int crtc_videochange (void)
{
 if ((hdisp) && (vdisp) && (scans_per_row))
    {
     if (hdisp < 50)		// for programs like Videotex (40 Col)
        crtc_y_scale = 1;
     else
        crtc_y_scale = crtc_y_scale_conf;        

     gui_videochange(hdisp * 8, vdisp * scans_per_row);

     crtc_SetRectValues();
     SDL_FreeSurface(vdu); 
     
     if (crtc_CreateRGBSurface(hdisp * 8, vdisp * scans_per_row) == -1)
        return -1;

     crtc_setmonitor();
    
     crtc_set_redraw();
     gui_redraw();

     crtcresized = 0;		// clear the resized flag
    }
 return 0;
} 

//==============================================================================
// CRTC Clock method - determines the VBLANK rate and method to be used. The
// method determines if XTAL frequency or host timer will be used for VBLANK
// calculations.
//
// Application software was designed for 2 and 3.75 MHz XTALS and the VBLANK
// can be made to emulate those two clock rates at a nominal 50Hz.
//
// The 6545 registers are not used at present for this calculation.
//
// While method 1 may appear to be the obvious choice here for determining
// the VBLANK period it is not the best choice for normal emulation speed
// because the host timer does not stop counting but the Z80 cycle count
// does stop after the required block of Z80 cycles has executed and so will
// then see all the the VBLANK states.
//
//   pass: int cpuclock		XTAL frequency in Hz.
//         int method		0=50 Hz VBLANK is calculated using Z80 cycles
//                                   and is proportional to the XTAL frequency.
//				1=50 Hz VBLANK is determined by host timer.
// return: void
//==============================================================================
void crtc_clock (int cpuclock, int method)
{
 vblank_method = method;
 if (vblank_method == 0)
    vblank_divval = cpuclock / 100; // 33750 if 3.375 MHz
}

//==============================================================================
// CRTC Initialise
//
//   pass: void
// return: int			0 if sucess, -1 if error
//==============================================================================
int crtc_init (void)
{
 if (crtc_CreateRGBSurface(VDUWIDTH, VDUHEIGHT) == -1)
    return -1;

 crtc_setmonitor();

 return 0;
}

//==============================================================================
// CRTC de-initialise
//
//   pass: void
// return: int			0
//==============================================================================
int crtc_deinit (void)
{
 SDL_FreeSurface(vdu);
 
 return 0;
}

//==============================================================================
// CRTC reset
//
//   pass: void
// return: int			0
//==============================================================================
int crtc_reset (void)
{
 reg = 0;

 return 0;
}

//==============================================================================
// CRTC vblank status
//
// The Vertical blanking status is generated from the Z80 clock cycles that
// have elapsed or the host timer depending on the mode required.
//
// The Vertical blanking period is emulated to produce a 50% duty cycle, on
// real hardware the blanking on time would be a much smaller period, but I
// have not found any software where this is a problem.  The emulated
// blanking frequency is 50 frames per second for normal useage or can be
// a proportional value calculated from the XTAL frequency.
//
// The vertical blanking is commonly used for keyboard encoding delays and for
// delays used in some games.
//
//   pass: void
// return: int			vblank status (in bit 7)
//==============================================================================
int crtc_vblank (void)
{
 uint64_t cycles_now;

 if (vblank_method == 0)
    {
     cycles_now = z80_cycles + mz80GetElapsedTicks(0); // use 0 !
     if ((cycles_now / vblank_divval) & 0x01)
        return B8(10000000);
    }
 else
    {
#ifdef MINGW
     if ((clock() / 10) & 0x01)
        return B8(10000000);
#else
     static struct timeval todx;
     gettimeofday(&todx, NULL);          
     if ((todx.tv_usec / 10000) & 0x01) // div by 100mS (rolls over each sec)
        return B8(1000000);
#endif
    }       
 return 0;
}

//==============================================================================
// CRTC status
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16		status
//==============================================================================
UINT16 crtc_status (UINT16 port, struct z80PortRead *port_s)
{
 int status = 0x80;                    	// always set update strobe

 if ((modelx.lpen) && (! lpen_valid))
    keystd_checkall();

 if ((modelx.lpen) && (lpen_valid))     // NB: this is not an else because
    status |= 0x40;                     // keystd_checkall might set lpen_valid

 if (crtc_vblank())
    status |= 0x20;

 if (debuglvl > 0)
    fprintf(debuglog, "crtc_status read = 0x%02x\n", status);

 return status;
}

//==============================================================================
// CRTC Light Pen
//
// Thi is only called by non Teleterm models.
//
//   pass: int addr
// return: void
//==============================================================================
void crtc_lpen (int addr)
{
 if (! lpen_valid)
    {
     lpen_valid = 1;
     lpen = addr;
     if (debuglvl > 0)
        fprintf(debuglog, "crtc_lpen addr = 0x%04x\n", addr);
    }
}

//==============================================================================
// CRTC set reg address - Port function 
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void crtc_address (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 reg = data & 0x1F;

// if (debuglvl > 0)
//    fprintf(debuglog, "crtc_address set = 0x%02x\n", data);
}

//==============================================================================
// CRTC read register data - Port function
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16 register 	register data
//==============================================================================
UINT16 crtc_data_r (UINT16 port, struct z80PortRead *port_s)
{
// if (debuglvl > 0)
//    fprintf(debuglog, "crtc_data read\n");

 switch (reg)
    {
     case CRTC_CUR_POS_H:
        return (cur_pos >> 8) & 0x3F;
        break;
     case CRTC_CUR_POS_L:
        return cur_pos & 0xFF;
        break;

     case CRTC_LPEN_H:
        if (debuglvl > 0)
           fprintf(debuglog, "crtc_lpen_h: lpen = 0x%04x\n", lpen);
        lpen_valid = 0;
        return (lpen >> 8) & 0x3F;
        break;
     case CRTC_LPEN_L:
        lpen_valid = 0;
        if (debuglvl > 0)
           fprintf(debuglog, "crtc_lpen_l: lpen = 0x%04x\n", lpen);
        return lpen & 0xFF;
        break;

     default:
        return 0xFFFF;
   }
}

//==============================================================================
// CRTC write register data - Port function
//
// crtc_redraw, and crtc_redraw_char functions
// -------------------------------------------
// uses: hdisp, vdisp, disp_start, scans_per_row, cur_start, cur_end,
//       cur_pos, cur_mode;
//
// Setting up the display resolution
// ---------------------------------
// The CRTC_HDISP data is placed in hdisp.
// The CRTC_VDISP data is placed in vdisp.
// The CRTC_SCANLINES data is placed in scans_per_row
//
// The X resolution is determined from hdisp * 8 The Y resolution is
// determined from vdisp * scans_per_row
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void crtc_data_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 int old_curpos;

// if (debuglvl > 0)
//    fprintf(debuglog, "crtc_data_w: reg=%d data=%d\n", reg, data);

#if 0
 if (reg < 18)
    printf("crtc_data_w: reg=%d data=%d\n", reg, data);
#endif
 
 switch (reg)
    {
     case CRTC_HTOT:		// R0
        htot = (data & 0xFF) + 1;
        break;
     case CRTC_HDISP:		// R1
        if (hdisp != (data & 0xFF))	// if the value has really changed
           {
            hdisp = data & 0xFF;
            crtcresized = 1;
           }
        break;

     case CRTC_VTOT:		// R4
        vtot = (data & 0x7F) + 1;
        break;
     case CRTC_VTOT_ADJ:	// R5
        vtot_adj = data & 0x1F;
        break;

     case CRTC_VDISP:		// R6
        if (vdisp != (data & 0x7F))	// if the value has really changed
           {
            vdisp = data & 0x7F;
            crtcresized = 1;
           }
        break;

     case CRTC_SCANLINES:	// R9
        if (scans_per_row != ((data & 0x1F) + 1)) // if the value has really changed
           {
            scans_per_row = (data & 0x1F) + 1;
            crtcresized = 1;
           }
        break;

     case CRTC_CUR_START:	// R10
        cur_start = data & 0x1F;
        cur_mode = (data >> 5) & 0x03;
        crtc_redraw_char(cur_pos);
        break;
     case CRTC_CUR_END:		// R11
        cur_end = data & 0x1F;
        crtc_redraw_char(cur_pos);
        break;

     case CRTC_DISP_START_H:	// R12
        disp_start &= 0xFF;
        disp_start |= (data & 0x3F) << 8;
        break;
     case CRTC_DISP_START_L:	// R13
        disp_start &= 0x3F00;
        disp_start |= data & 0xFF;
        break;

     case CRTC_CUR_POS_H:	// R14
        old_curpos = cur_pos;
        cur_pos &= 0xFF;
        cur_pos |= (data & 0x3F) << 8;
        crtc_redraw_char(old_curpos);
        crtc_redraw_char(cur_pos);
        break;
     case CRTC_CUR_POS_L:	// R15
        old_curpos = cur_pos;
        cur_pos &= 0x3F00;
        cur_pos |= data & 0xFF;
        crtc_redraw_char(old_curpos);
        crtc_redraw_char(cur_pos);
        break;

     case CRTC_SETADDR_H:	// R18
        mem_addr = (mem_addr & 0x00FF) | ((data & 0x3F) << 8);
        break;
     case CRTC_SETADDR_L:	// R19
        mem_addr = (mem_addr & 0x3F00) | (data & 0xFF);
        break;
        
     case CRTC_DOSETADDR:	// R31
        if (debuglvl > 0)
            fprintf(debuglog, "crtc_dosetaddr: mem_addr = 0x%04x\n", mem_addr);
        keystd_handler(mem_addr);
        break;
   }
}

//==============================================================================
// CRTC redraw one screen address character position.
//
// Note: If the 'standard' option is in affect then this code should behave like
// a standard machine. (extendram=0, attribram=0)
//
//   pass: int addr		address to be redrawn
// return: void
//==============================================================================
void crtc_redraw_char (int addr)
{
 int i, j, x, y, offset, maddr;
 char ch, cursor, inverse;
 char *data;

 Uint32 xfgc;
 Uint32 xbgc;

 UINT8 colour;
 
 SDL_Rect src, dst;

 int pcgbank;
 char attrib; 
 int fgc, bgc;

 if (hdisp == 0)
    return;

 offset = (addr & 0x07FF) - (disp_start & 0x07FF);  // ofs from disp_start
 if (offset < 0)
    offset += 0x0800;
 x = offset % hdisp;			// character column
 y = offset / hdisp;			// character row
 maddr = (disp_start + offset) & 0x3FFF;

 ch = screen_ram_ap[maddr & 0x07FF];
 inverse = 0;
 
 if (extendram)
    {
     attrib = attrib_ram_ap[maddr & 0x07FF];
     pcgbank = attrib & 0x0F;

     if (attrib & B8(11000000))	        // if inverse or flash attribute bit set
        {
         if (attrib & B8(01000000))	// if inverse attribute bit set
            inverse = 0xFF;
         if (attrib & B8(10000000))	// if flash attribute bit set                
            if (flashvideo)
               ch = 32;			// use blank instead when flashing
        }     
     
     if (ch & 0x80)      		// selects PCG RAM
        data = pcg_ram_ap + pcgbank * 0x0800 + 16 * (ch & 0x7F);
     else
        data = charrom + 16 * (ch & 0x7F) + ((maddr & 0x2000) >> 2);
    }
 else
    {
     if (ch & 0x80)      		// selects PCG RAM
        data = pcg_ram_ap + 16 * (ch & 0x7F);
     else
        data = charrom + 16 * (ch & 0x7F) + ((maddr & 0x2000) >> 2);
    } 

 if (! monochrome)
    {
     colour = colour_ram_ap[maddr & 0x07FF];
     fgc = (colour & 0x0F);
     bgc = (colour >> 4);
     xfgc = SDL_MapRGB(spf, col_table[fgc][0], col_table[fgc][1], col_table[fgc][2]);
     xbgc = SDL_MapRGB(spf, col_table[bgc][0], col_table[bgc][1], col_table[bgc][2]);
    }
 else
    {
     xfgc = fgcolour;   
     xbgc = bgcolour;
    } 

 if ((x >= hdisp) || (y >= vdisp))
    return;

 SDL_LockSurface(vdu);
 for (i=0; i < scans_per_row; i++)
    {
     if ((maddr == cur_pos) && (cur_mode != 1))
        {
         if ((i >= cur_start) && (i <= cur_end))
            {
             if (cur_mode == 0)
                cursor = 0xFF;
             else
                cursor = cur_blink;
            }
         else
            cursor = 0x00;    
        }   
     else
        cursor = 0x00;

     ch = *(data++) ^ inverse ^ cursor;

     for (j=0; j < 8; j++)
        {
         if (ch & 0x80)
            {
             if (crtc_y_scale == 2)
                {
                 putpixel(vdu, x*8+j, y*crtc_y_scale*scans_per_row+(i*2), xfgc);
                 putpixel(vdu, x*8+j, y*crtc_y_scale*scans_per_row+(i*2)+1, xfgc);
                }
             else    
                putpixel(vdu, x*8+j, y*scans_per_row+i, xfgc);
            }    
         else
            {
             if (crtc_y_scale == 2)
                {
                 putpixel(vdu, x*8+j, y*crtc_y_scale*scans_per_row+(i*2), xbgc);
                 putpixel(vdu, x*8+j, y*crtc_y_scale*scans_per_row+(i*2)+1, xbgc);
                }
             else
                putpixel(vdu, x*8+j, y*scans_per_row+i, xbgc);
            }    
         ch <<= 1;
        }
    }
 SDL_UnlockSurface(vdu);

 /* do sub surfaces ?!? */
 src.x = x * 8;
 src.y = y * scans_per_row * crtc_y_scale;
 src.w = 8;
 src.h = scans_per_row * crtc_y_scale;
 dst.x = VDUPOSX + src.x;
 dst.y = VDUPOSY + src.y;
 SDL_BlitSurface(vdu, &src, screen, &dst);
}

//==============================================================================
// CRTC set the redraw flag so that the next crtc_redraw function call is
// carried out.
//
//   pass: void
// return: void
//==============================================================================
void crtc_set_redraw (void)
{
 redraw = 1;
}

//==============================================================================
// CRTC update the whole screen area if redraw is non 0.
//
// Note: If the 'standard' option is in affect then this code should behave like
// a standard machine. (extendram=0, attribram=0)
//
//   pass: void
// return: void
//==============================================================================
void crtc_redraw (void)
{
 int i, j, k, l;
 int maddr;
 char ch, cursor, inverse;
 char *data;

 Uint32 xfgc;
 Uint32 xbgc;

 UINT8 colour;

 int pcgbank;
 char attrib;
 int fgc, bgc;

 if (! redraw)
    return;

 redraw = 0;

 SDL_LockSurface(vdu);
 maddr = disp_start;

 for (i=0; i<vdisp; i++)
    for (j=0; j<hdisp; j++)
       {
        ch = screen_ram_ap[maddr & 0x07FF];
        inverse = 0;
        
        if (extendram)
           {
            attrib = attrib_ram_ap[maddr & 0x07FF];
            pcgbank = attrib & 0x0F;

            if (attrib & B8(11000000))	// if inverse or flash attribute bit set
               {
                if (attrib & B8(01000000))  // if inverse attribute bit set
                   inverse = 0xFF;
                if (attrib & B8(10000000))  // if flash attribute bit set                
                   if (flashvideo)
                      ch = 32;		// use blank instead when flashing
               }
               
            if (ch & 0x80)      	// if PCG RAM
               data = pcg_ram_ap + pcgbank * 0x0800 + 16 * (ch & 0x7F);
            else
               data = charrom + 16 * (ch & 0x7F) + ((maddr & 0x2000) >> 2);
           }
        else
           {
            if (ch & 0x80)      	// if PCG RAM
               data = pcg_ram_ap + 16 * (ch & 0x7F);
            else
               data = charrom + 16 * (ch & 0x7F) + ((maddr & 0x2000) >> 2);
           } 

        if (! monochrome)
           {   
            colour = colour_ram_ap[maddr & 0x07FF];
            fgc = (colour & 0x0F);
            bgc = (colour >> 4);
            xfgc = SDL_MapRGB(spf, col_table[fgc][0], col_table[fgc][1], col_table[fgc][2]);
            xbgc = SDL_MapRGB(spf, col_table[bgc][0], col_table[bgc][1], col_table[bgc][2]);
           } 
        else
           {
            xfgc = fgcolour;
            xbgc = bgcolour;
           } 

        for (k=0; k < scans_per_row; k++)    
           {
            if ((maddr == cur_pos) && (cur_mode != 1))
               {
                if ((k >= cur_start) && (k <= cur_end))
                   {
                    if (cur_mode == 0)
                       cursor = 0xFF;
                    else
                       cursor = cur_blink;
                   }
                else
                   cursor = 0x00;    
               }   
            else
               cursor = 0x00;

            ch = *(data++) ^ inverse ^ cursor;

            for (l=0; l < 8; l++)
               {
                if (ch & 0x80)
                   {
                    if (crtc_y_scale == 2)
                       {
                        putpixel(vdu, j*8+l, i*crtc_y_scale*scans_per_row+(k*2), xfgc);
                        putpixel(vdu, j*8+l, i*crtc_y_scale*scans_per_row+(k*2)+1, xfgc);
                       }
                    else    
                       putpixel(vdu, j*8+l, i*scans_per_row+k, xfgc);
                   }    
                else
                   {
                    if (crtc_y_scale == 2)
                       {
                        putpixel(vdu, j*8+l, i*crtc_y_scale*scans_per_row+(k*2), xbgc);
                        putpixel(vdu, j*8+l, i*crtc_y_scale*scans_per_row+(k*2)+1, xbgc);
                       }
                    else
                       putpixel(vdu, j*8+l, i*scans_per_row+k, xbgc);
                   }    
                ch <<= 1;
               }
           }

        maddr = (maddr + 1) & 0x3FFF;
       }

 SDL_UnlockSurface(vdu);
 SDL_BlitSurface(vdu, &rect_vdu, screen, &rect_screen);
}

//==============================================================================
// crtc configure.
//
// Determines the displayed aspect ratio to use.
//
//   pass: int aspect
// return: void
//==============================================================================
void crtc_configure (int aspect)
{
 crtc_y_scale_conf = aspect; 
 crtc_y_scale = aspect;
}
