//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                              WD2793 FDC module                             *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Emulates the WD2793 floppy disk controller.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.1.0 - 24 October 2007, S.J.Kay
// - The FDC_READADDR command now calls a new disk_read_idfield() function to
//   obtain the track ID field.  The previous code had assumed all sectors were
//   in the 1 to 10 range and would not work on the 256TC and the DS80 Modular
//   disk images,  the sector size also returns the size for the disk selected.
// - Changes made to allow disk formats that have sectors that start from
//   greater than 1 to be read. i.e DS80 Microbee disks are 10 sectors, with
//   sector numbers 21-30 on the data tracks.
// - Implement the modelx information structure.
//
// v2.0.0 - 7 October 2007, S.J.Kay
// - Each model emulated now has a preferred boot image.  If that image can not
//   be opened then the fallback image is used.
// - Added fdc_status0 variable for a --nodisk option.  This returns a status
//   value of 0 from the fdc_ext_r() function, and is used to simulate a no
//   disk in drive condition.
//
// v1.4.0 - 26 September 2007, S.J.Kay
// - Changes to ignore functions if a ROM model is being emulated.
// - Added changes to error reporting, replaced exit(1) with return -1 in
//   fdc_loaddisk() function.  
//
// v1.3.0 - 1 September 2007, S.J.Kay
// - Changes some error messages and redirection.
//
// v1.0.0 - 2 July 2007, S.J.Kay
// - Implement DSK disk image format instead of raw images.
// - All references to '10' sectors per track have been replaced by
//   drive->disk->imagerec.sectrack and references to '512' have been replaced
//   by drive->disk->imagerec.secsize.
// - Disk images are no longer assumed to MAXTRACK=39,  for now change to
//   255 (largest value allowed for WD2793 ?)
// - BUFSSZE increased from 10x512 to 10x1024 byte sectors.
// - Added 2 more drives (now is A-D).  The drive images can be specified on
//   the command line. If drive 'A' is not specified then "boot.dsk" will be
//   used by default.  Drives 'B-D' must be specified if required.
// - implemented case FDC_WRITESECT: in fdc_cmd function.
// - implemented fdc_data_w function.
// - in fdc_data_r changed the ctrl_status value at end sector read.  The
//   value of FDC_WRPROT did not look correct here.
// - Added a second drive.  
//
// v0.0.0 - 5 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>

#include "fdc.h"
#include "disk.h"
#include "ubee512.h"

static int fdc_loaddisk (const char *name, int drive, int report);
static void fdc_unloaddisk (int drive);
static int fdc_bootimage (void);

//==============================================================================
// structures and variables
//==============================================================================
static int ctrl_side, ctrl_drive, ctrl_ddense;
static int ctrl_rdata, ctrl_rtrack, ctrl_rsect,
           ctrl_intrq, ctrl_drq, ctrl_status,
           ctrl_stepdir;


static fdc_drive_t fdc_drive[FDC_NUMDRIVES];

static char buf[FDC_BUFSIZE];
static int bytes_left, buf_index;

int fdc_status0;

extern int model;
extern model_t modelx;

extern char userhome_diskpath[];
extern char *model_args[];

extern char floppya[];
extern char floppyb[];
extern char floppyc[];
extern char floppyd[];

extern FILE *debuglog;
extern int debuglvl;

//==============================================================================
// Initialise
//
//   pass: void
// return: int			0 if no errors, else -1
//==============================================================================
int fdc_init (void)
{
 int i;
 int res;

 if (modelx.rom)
    return 0;

 for (i=0; i < FDC_NUMDRIVES; i++)
    fdc_drive[i].disk = NULL;

 res = fdc_bootimage();
 
 if (strlen(floppyb) && (res == 0))
    {
     printf("Image B: %s\n", floppyb);
     res = (fdc_loaddisk(floppyb, 1, 1));
    }    

 if (strlen(floppyc) && (res == 0))
    {
     printf("Image C: %s\n", floppyc);
     res = (fdc_loaddisk(floppyc, 2, 1));
    }    
     
 if (strlen(floppyd) && (res == 0))
    {
     printf("Image D: %s\n", floppyd);
     res = (fdc_loaddisk(floppyd, 3, 1));
    }    

 if (res < 0)
    return -1;
 return 0;   
}

//==============================================================================
// De-initialise
//
//   pass: void
// return: int			0
//==============================================================================
int fdc_deinit (void)
{
 int i;

 if (modelx.rom)
    return 0;

 for (i=0; i < FDC_NUMDRIVES; i++)
    fdc_unloaddisk(i);
 
 return 0;
}

//==============================================================================
// reset the controller
//
//   pass: void
// return: int			0
//==============================================================================
int fdc_reset (void)
{
 if (modelx.rom)
    return 0;

 ctrl_side = ctrl_drive = ctrl_ddense = 0;
 ctrl_rtrack = ctrl_intrq = ctrl_drq = 0;
 ctrl_stepdir = +1;

 fdc_drive[ctrl_drive].track = 0;
 ctrl_status = FDC_TRACK0;

 if (fdc_drive[ctrl_drive].disk == NULL)
    ctrl_status |= FDC_INDEXPULSE;              /* set index pulse if no disk */
 if (disk_iswrprot(fdc_drive[ctrl_drive].disk))
    ctrl_status |= FDC_WRPROT;
 
 bytes_left = 0;

 return 0;
}

//==============================================================================
// load disk
//
//   pass: const char *name
//         int drive
//         int report		1 if require error message if file not opened
// return: int			0 if no errors,  -1 if malloc error, else -2
//==============================================================================
static int fdc_loaddisk (const char *name, int drive, int report)
{
 fdc_unloaddisk(drive);

 fdc_drive[drive].disk = malloc(sizeof(disk_t));
 
 if (fdc_drive[drive].disk == NULL)
    {
     console_output();
     printf("fdc_loaddisk: Unable to malloc");
     return -1;
    }

 if (disk_open(name, fdc_drive[drive].disk) != 0)
    {
     fdc_unloaddisk(drive);
     if (report)
        {
         console_output();
         printf("fdc_loaddisk: Unable to open %s\n", name);
         printf("fdc_loaddisk: Check image is not set as read only!\n");
        }
     return -2;
    }

 return 0;
}

//==============================================================================
// un-load disk
//
//   pass: int drive		drive number
// return: void
//==============================================================================
static void fdc_unloaddisk (int drive)
{
 if (fdc_drive[drive].disk != NULL)	// if malloc allocated
    {
     if (fdc_drive[drive].disk->fdisk)  // if opened
        disk_close(fdc_drive[drive].disk);
     free(fdc_drive[drive].disk);
     fdc_drive[drive].disk = NULL;
    } 
}

//==============================================================================
// Load boot image.
//
// Load the default boot disk for the model being emulated.  if the default
// boot image is not found then the fall-back image will be booted.
//
//   pass: 
// return: int			0 if no errors,  -1 if malloc error, else -2
//==============================================================================
static int fdc_bootimage (void)
{
 char name[512];
 int res;
 
// if an image for drive A: was specified then load that one
 if (strlen(floppya))
    {
     printf("Image A: %s\n", floppya);
     res = (fdc_loaddisk(floppya, 0, 1));
    }    
 else
// load the default image for the model being emulated
    {   
     snprintf(name, sizeof(name), "%s%s.dsk", userhome_diskpath, model_args[model]);
     printf("Image A: %s\n", name);
     res = (fdc_loaddisk(name, 0, 0));

// if the default image was not found then use the fallback boot image
     if (res == -2)
        {
         snprintf(name, sizeof(name), "%s"BOOT_IMAGE, userhome_diskpath);
         printf("Image not able to be opened, trying fall-back boot image\n");
         printf("Image A: %s\n", name);
         res = (fdc_loaddisk(name, 0, 1));
        }
    }

 return res;
}

//==============================================================================
// FDC command
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_cmd (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 int cmd;
 int sectoffs;

 read_addr_t *idfield = (read_addr_t *)buf;

 if (modelx.rom)
    return;

 fdc_drive_t *drive = &fdc_drive[ctrl_drive];

 if (drive->disk == NULL)
    {
     ctrl_status = FDC_RECNOTFOUND;
     return;
    } 

 ctrl_intrq = 0;  // new command clears interrupt, FIXME
    
 if (ctrl_rsect > drive->disk->imagerec.sectrack)
    sectoffs = drive->disk->imagerec.datasecofs - 1;
 else
    sectoffs = 0;

 cmd = (data >> 4) & 0x0F;  // get command bits
 if ( ((cmd >> 1) >= 1) && ((cmd >> 1) <= 5) )
    cmd &= 0x0E;  // get rid of data bits in the command field

 switch (cmd)
    {
     case FDC_RESTORE:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: restore\n");

        drive->track = ctrl_rtrack = 0;

        ctrl_intrq = 1;
        ctrl_status = FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(drive->disk))
           ctrl_status |= FDC_WRPROT;
        if (drive->disk == NULL)
           ctrl_status |= FDC_INDEXPULSE;
        break;

     case FDC_SEEK:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: seek\n");

        drive->track += ctrl_rdata - ctrl_rtrack;
        if (drive->track < 0)
           drive->track = 0;
        else if (drive->track > FDC_MAXTRACK)
           drive->track = FDC_MAXTRACK;

        if (ctrl_rdata < ctrl_rtrack)
           ctrl_stepdir = -1;
        else
           ctrl_stepdir = +1;
        ctrl_rtrack = ctrl_rdata;

        ctrl_intrq = 1;
        ctrl_status = 0;
        if (drive->track == 0)
           ctrl_status |= FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(drive->disk))
           ctrl_status |= FDC_WRPROT;
        if (drive->disk == NULL)
           ctrl_status |= FDC_INDEXPULSE;
        break;

     case FDC_STEP:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: step\n");

        drive->track += ctrl_stepdir;
        if (drive->track < 0)
           drive->track = 0;
        else
           if (drive->track > FDC_MAXTRACK)
              drive->track = FDC_MAXTRACK;

        if (data & FDC_UPDATETRACK)
           ctrl_rtrack += ctrl_stepdir;

        ctrl_intrq = 1;
        ctrl_status = 0;
        if (drive->track == 0)
           ctrl_status |= FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(drive->disk))
           ctrl_status |= FDC_WRPROT;
        if (drive->disk == NULL)
           ctrl_status |= FDC_INDEXPULSE;
        break;

     case FDC_STEPIN:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: stepin\n");

        drive->track++;
        if (drive->track > FDC_MAXTRACK)
           drive->track = FDC_MAXTRACK;

        if (data & FDC_UPDATETRACK)
           ctrl_rtrack++;

        ctrl_stepdir = +1;

        ctrl_intrq = 1;
        ctrl_status = 0;
        if (drive->track == 0)
           ctrl_status |= FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(drive->disk))
           ctrl_status |= FDC_WRPROT;
        if (drive->disk == NULL)
           ctrl_status |= FDC_INDEXPULSE;
        break;

     case FDC_STEPOUT:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: stepout\n");

        drive->track--;
        if (drive->track < 0)
           drive->track = 0;

        if (data & FDC_UPDATETRACK)
           ctrl_rtrack--;

        ctrl_stepdir = -1;

        ctrl_intrq = 1;
        ctrl_status = 0;
        if (drive->track == 0)
           ctrl_status |= FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(drive->disk))
           ctrl_status |= FDC_WRPROT;
        if (drive->disk == NULL)
           ctrl_status |= FDC_INDEXPULSE;
        break;

     case FDC_READSECT:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: readsect\n");

        if ((drive->disk == NULL) || (ctrl_rsect < 1) || (ctrl_rsect > (drive->disk->imagerec.sectrack + sectoffs)))
           {
            ctrl_status = FDC_RECNOTFOUND;
            break;
           }

        disk_read(drive->disk, buf, drive->disk->imagerec.secsize, ctrl_side, drive->track, ctrl_rsect);        
        buf_index = 0;
        bytes_left = drive->disk->imagerec.secsize;        

        if (data & FDC_MULTISECT)
           {
            while (++ctrl_rsect <= drive->disk->imagerec.sectrack + sectoffs)
               {
                disk_read(drive->disk, &buf[bytes_left], drive->disk->imagerec.secsize, ctrl_side, drive->track, ctrl_rsect);               
                bytes_left += drive->disk->imagerec.secsize;               
               }
           }    

        ctrl_status = FDC_BUSY | FDC_DRQ;
        ctrl_drq = 1;
        break;

     case FDC_WRITESECT:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: writesect\n");

        if ((drive->disk == NULL) || (ctrl_rsect < 1) || (ctrl_rsect > (drive->disk->imagerec.sectrack + sectoffs)))
           {
            ctrl_status = FDC_RECNOTFOUND;
            break;
           }

        buf_index = 0;
        bytes_left = drive->disk->imagerec.secsize;        

        if (data & FDC_MULTISECT)
           while (++ctrl_rsect <= drive->disk->imagerec.sectrack + sectoffs)
              bytes_left += drive->disk->imagerec.secsize;

        ctrl_status = FDC_BUSY | FDC_DRQ;
        ctrl_drq = 1;
        break;

     case FDC_READADDR:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: readaddr\n");

        bytes_left = 6;
        buf_index = 0;
        disk_read_idfield(drive->disk, idfield, ctrl_side, drive->track);
        ctrl_rsect = idfield->track;

        ctrl_status = FDC_BUSY | FDC_DRQ;
        ctrl_drq = 1;
        break;

     case FDC_READTRACK:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: readtrack\n");
        break;

     case FDC_WRITETRACK:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: writetrack\n");
        break;

     case FDC_INTERRUPT:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: interrupt\n");

        ctrl_status = FDC_WRPROT;                        /* FIXME: check this */
        bytes_left = 0;
        ctrl_drq = 0;
        ctrl_intrq = 1;
        break;
    }
}

//==============================================================================
// Read FDC status
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			ctrl_status
//==============================================================================
UINT16 fdc_status (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.rom)
    return 0;

 ctrl_intrq = 0;                     /* read status clears interrupt, FIXME for
                                                    immediate interrupt case */

 if (bytes_left != 0)                                       /* FIXME: timing */
    {
     buf_index++;
     ctrl_status |= FDC_LOSTDATA;
     if (--bytes_left == 0)
        {
         ctrl_status = FDC_WRPROT | FDC_LOSTDATA;                  /* FIXME! */
         ctrl_drq = 0;
         ctrl_intrq = 1;
        }
    }
 
 if (debuglvl > 0)
    fprintf(debuglog, "fdc_status: port=0x%02X crtl_status=0x%02X\n", port, ctrl_status);
    
// printf("fdc_status: port=0x%02X crtl_status=0x%02X\n", port, ctrl_status);
 return ctrl_status;
}

//==============================================================================
// Write 1 data byte to sector
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_data_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.rom)
    return;

 fdc_drive_t *drive = &fdc_drive[ctrl_drive];

 if (bytes_left != 0)
    {
     buf[buf_index++] = data;
     if (--bytes_left == 0)            	// if end of sector write
        {
         disk_write(drive->disk, buf, drive->disk->imagerec.secsize, ctrl_side, drive->track, ctrl_rsect);         
         ctrl_drq = 0;
         ctrl_intrq = 1;
         ctrl_status &= 0xFF ^ (FDC_BUSY | FDC_DRQ );
        }
    }

 if (debuglvl > 0)
    fprintf(debuglog, "fdc_data_w: port=0x%02X crtl_rdata=0x%02X\n", port, data);

 ctrl_rdata = data;
}

//==============================================================================
// Read 1 data byte from sector
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: UINT16 			ctrl_rdata
//==============================================================================
UINT16 fdc_data_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.rom)
    return 0;

 if (bytes_left != 0)
    {
     ctrl_rdata = buf[buf_index++];
     if (--bytes_left == 0)		// if end of sector read
        {
         ctrl_drq = 0;
         ctrl_intrq = 1;
         ctrl_status &= 0xFF ^ (FDC_BUSY | FDC_DRQ );         
        }
    }

 if (debuglvl > 0)
    fprintf(debuglog, "fdc_data_r: port=0x%02X crtl_rdata=0x%02X\n", port, ctrl_rdata & 0xFF);

 return ctrl_rdata & 0xFF;
}

//==============================================================================
// Set track number
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_track_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.rom)
    return;

 if (debuglvl > 0)
    fprintf(debuglog, "fdc_track_w: port=0x%02X ctrl_rtrack=0x%02X\n", port, data);
#if 0
 printf("fdc_track_w: port=0x%02X ctrl_rtrack=0x%02X\n", port, data);
#endif
 ctrl_rtrack = data;
}

//==============================================================================
// Get track number
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: UINT16			ctrl_rtrack
//==============================================================================
UINT16 fdc_track_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.rom)
    return 0;

 if (debuglvl > 0)
    fprintf(debuglog, "fdc_track_r: port=0x%02X ctrl_rtrack=0x%02X\n", port, ctrl_rtrack);
#if 0
 printf("fdc_track_r: port=0x%02X ctrl_rtrack=0x%02X\n", port, ctrl_rtrack);
#endif
 return ctrl_rtrack;
}

//==============================================================================
// Set sector number
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_sect_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.rom)
    return;

 if (debuglvl > 0)
    fprintf(debuglog, "fdc_sect_w: port=0x%02X ctrl_rsect=0x%02X\n", port, data);

 ctrl_rsect = data;
}

//==============================================================================
// Get sector number
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: UINT16			ctrl_rsect
//==============================================================================
UINT16 fdc_sect_r (UINT16 port, struct z80PortRead *port_s)
{
 if (modelx.rom)
    return 0;

 if (debuglvl > 0)
    fprintf(debuglog, "fdc_sect_r: port=0x%02X ctrl_rsect=0x%02X\n", port, ctrl_rsect);

 return ctrl_rsect;
}

//==============================================================================
// Set drive, side and density
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_ext_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.rom)
    return;

 if (debuglvl > 0)
    fprintf(debuglog, "fdc_ext_w: port=0x%02X ctrl_drive/side/ddense=0x%02X\n", port, data);

 ctrl_drive = data & 0x03;
 ctrl_side = (data & 0x04) >> 2;
 ctrl_ddense = (data & 0x08) >> 3;
}

//==============================================================================
// Get intrq and drq
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0x80 if intrq or drq set, else 0
//==============================================================================
UINT16 fdc_ext_r (UINT16 port, struct z80PortRead *port_s)
{
 int status;

 if (modelx.rom)
    return 0;

 if (ctrl_intrq | ctrl_drq) 
    status = 0x80;
 else
    status = 0x00;
 
 if (debuglvl > 0)
    fprintf(debuglog, "fdc_ext_r: port=0x%02X (ctrl_intrq | crtl_drq) status=0x%02X\n", port, status);

// printf("fdc_ext_r: port=0x%02X (ctrl_intrq | crtl_drq) status=0x%02X\n", port, status);

// force returning a status value of 0,  used to simulate no disk in drive
 if (fdc_status0)
    {
     fdc_status0--;
     return 0;
    }

 return status;
}
