//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                              function module                               *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Built in functions able to be called from Z80 code.
// NO ACTUAL FUNCTIONALLITY IS CURRENTLY IMPLEMENTED.
//
// General purpose functions that don't belong to any particular module.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.4.0 - 22 September 2007, S.J.Kay
// - Added copy_file() function.
//
// v1.2.0 - 19 August 2007, S.J.Kay
// - Created new module to implement the emulator function ports.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#ifdef MINGW
#include <time.h>
#else
#include <sys/times.h>
#endif

#include "function.h"
#include "ubee512.h"
#include "mz80.h"

static const char emulator_id_str[] = APPIDSTR;  // defined in the Makefile
static int emulator_id_pos;

static int ubee_req;
static int ubee_addr;
static int ubee_command;

extern char z80mem[];

extern FILE *debuglog;
extern int debuglvl;

//==============================================================================
// Copy a file from one location to another.
//
//   pass: char *dest			destination file path and name
//         char *src			source file path and name
// return: int				0 if no error, else -1
//==============================================================================
int copy_file (char *dest, char *src)
{
 FILE *srcfp;
 FILE *destfp; 
 char *buf[0x4000];
 int len;

 srcfp = fopen(src, "rb");
 if (srcfp == NULL)
    return -1;

 destfp = fopen(dest, "wb"); 
 if (destfp == NULL)
    {
     fclose(srcfp);		// close the source file
     return -1;
    }

 while (! feof(srcfp))
    {
     len = fread(buf, 1, sizeof(buf), srcfp);
     if (len)
        fwrite(buf, 1, len, destfp);     
    } 
 
 fclose(srcfp);
 fclose(destfp); 

 return 0;
}

//==============================================================================
// Test if system uses Big Endian or Little Endian format.
//
//   pass: void
// return: int				0 if Big Endian, else Little Endian
//==============================================================================
static int IsBigEndian (void)
{
#if defined(__aix__)
 return _true;
#else
 int i = 1;
 int c = *((char*)&i);
 return c == 0;
#endif
}

//==============================================================================
// Swap Endianess. Only works for even number variables sizes.
//
//   pass: char *n
//         int size			size in bytes of the variable
// return: void
//==============================================================================
void swap_endian (char *n, int size) 
{
 int i;
 int x;

 size /= 2;
 
 for (i = 0; i < size; i++)
    {
     x = *(n+i);
     *(n+i) = *(n+(size-1)-i);
     *(n+(size-1-i)) = x;
    } 
}

//==============================================================================
// Little Endian functions to swap byte order if on a Big Endian host.
//
//   pass: xintxx_t n
// return: xintxx_t			litle endian format
//==============================================================================
int16_t little_endian_16 (int16_t n)
{
 if (IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}
          
uint16_t little_endian_u16 (uint16_t n)
{
 if (IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}

int32_t little_endian_32 (int32_t n) 
{
 if (IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}
                                              
uint32_t little_endian_u32 (uint32_t n)
{
 if (IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}

int64_t little_endian_64 (int64_t n) 
{
 if (IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}

uint64_t little_endian_u64 (uint64_t n) 
{
 if (IsBigEndian())
    swap_endian ((char *)&n, sizeof(n));
 return(n);
}

//==============================================================================
// Big Endian functions to swap byte order if on a Little Endian host.
//
//   pass: xintxx_t n
// return: xintxx_t			big endian format
//==============================================================================
int16_t big_endian_16 (int16_t n)
{
 if (! IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}
          
uint16_t big_endian_u16 (uint16_t n)
{
 if (! IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);    
}

int32_t big_endian_32 (int32_t n) 
{
 if (! IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}
                                              
uint32_t big_endian_u32 (uint32_t n)
{
 if (! IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}

int64_t big_endian_64 (int64_t n) 
{
 if (! IsBigEndian())
    swap_endian((char *)&n, sizeof(n));
 return(n);
}

uint64_t big_endian_u64 (uint64_t n) 
{
 if (! IsBigEndian())
    swap_endian ((char *)&n, sizeof(n));
 return(n);
}

//==============================================================================
// set Z80 memory 8, 16 and 32 bit signed and unsigned variables in little
// endian format.  If host uses big endian format the values are swapped first.
//
//   pass: int addr
//         xintxx_t  n
// return: void
//==============================================================================
static void z80_set_8 (int addr, int8_t n)
{
 z80mem[addr] = n;
}

static void z80_set_u8 (int addr, uint8_t n)
{
 z80mem[addr] = n;
}

static void z80_set_16 (int addr, int16_t n)
{
 n = little_endian_16(n);
 memcpy(z80mem+addr, &n, 2);
}

static void z80_set_u16 (int addr, uint16_t n)
{
 n = little_endian_u16(n);
 memcpy(z80mem+addr, &n, 2);
}

static void z80_set_32 (int addr, int32_t n)
{
 n = little_endian_32(n);
 memcpy(z80mem+addr, &n, 4);
}

static void z80_set_u32 (int addr, uint32_t n)
{
 n = little_endian_u32(n);
 memcpy(z80mem+addr, &n, 4);
}

static void z80_set_64 (int addr, int64_t n)
{
 n = little_endian_64(n);
 memcpy(z80mem+addr, &n, 8);
}

static void z80_set_u64 (int addr, uint64_t n)
{
 n = little_endian_u64(n);
 memcpy(z80mem+addr, &n, 8);
}

//==============================================================================
// get Z80 memory 8, 16 and 32 bit signed and unsigned variables in little
// endian format and convert to host endian format.
//
//   pass: int addr
// return: xintxx_t 			in host endian format
//==============================================================================
static int8_t z80_get_8 (int addr)
{
 return z80mem[addr];
}

static uint8_t z80_get_u8 (int addr)
{
 return z80mem[addr];
}

static int16_t z80_get_16 (int addr)
{
 int16_t n;
 memcpy(&n, z80mem+addr, 2);
 return (little_endian_16(n));
}

static uint16_t z80_get_u16 (int addr)
{
 uint16_t n;
 memcpy(&n, z80mem+addr, 2);
 return (little_endian_u16(n));
}

static int32_t z80_get_32 (int addr)
{
 int32_t n;
 memcpy(&n, z80mem+addr, 4);
 return (little_endian_32(n));
}

static uint32_t z80_get_u32 (int addr)
{
 uint32_t n;
 memcpy(&n, z80mem+addr, 4);
 return (little_endian_u32(n));
}

static int64_t z80_get_64 (int addr)
{
 int64_t n;
 memcpy(&n, z80mem+addr, 8);
 return (little_endian_64(n));
}

static uint64_t z80_get_u64 (int addr)
{
 uint64_t n;
 memcpy(&n, z80mem+addr, 8);
 return (little_endian_u64(n));
}
 
//==============================================================================
// Function Initialise
//
//   pass: void
// return: int			0
//==============================================================================
int function_init (void)
{
 ubee_req = 0;

 return 0;
}

//==============================================================================
// Function de-initialise
//
//   pass: void
// return: int			0
//==============================================================================
int function_deinit (void)
{
 return 0;
}

//==============================================================================
// Function reset
//
//   pass: void
// return: int			0
//==============================================================================
int function_reset (void)
{
 ubee_req = 0;
 return 0;
}

//==============================================================================
// uBee512 read - Special Ubee512 Port function
//
// This function allows a Z80 program to identify that it is running in the
// uBee512 emulator.  This is provided so that some extra functionality can
// be made taken advantage off and must be called and checked before using any
// special uBee512 write function calls to prevent unknown results on a real
// Microbee.
//
// This port call just returns the program name and version as a string one
// character at a time in a circular loop,  the APPIDSTR in the Makefile
// ultimately determines the name and version ID to be used.
//
// 'u', 'B', 'e', 'e' '5', '1', '2', ' ', 'v', '1', '0', '0', 0
//
// To use the port call (pascal example):
// {---------------------------------- CUT ------------------------------------}
// const
//    EMUPORT=$FF;	{ still need to decide on a safe port number }
// var
//    name, string : string[10];
//    value, error : integer;
//    c : char;
//
// vers := '';
// name := '';
//
// repeat until (port[EMUPORT] = 0);
//
// repeat
//    c := upcase(chr(port[EMUPORT]));
//    if (c <> ord(' ') then
//       name := name + c;
// until (c = ' ');
//
// repeat
//    c := chr(port[EMUPORT]);
//    if ((c <> #0) and (c in ['0'..'9'])) then
//       vers := vers + chr(c);
// until (c = #0);
//
// if (name = 'UBEE512') then
//    begin    
//       val(vers, value, error);
//       if ((value => 100) and (error = 0)) then
//          begin
//             writeln('uBee512 emulator version: ', value);   
//          end
//       else
//          writeln('Not the correct uBee512 emulator version!');   
//    end
// else
//    writeln('Not running on the uBee512 emulator!');   
// {---------------------------------- CUT ------------------------------------}
//    
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16		returns a null terminated string in a loop
//==============================================================================
UINT16 function_ubee_r (UINT16 port, struct z80PortRead *port_s)
{
 char c;

 c = emulator_id_str[emulator_id_pos++];
 if (! c)
    emulator_id_pos = 0;
 
#if 0
 printf("function_ubee_r port %c\n", c);
#endif

 if (debuglvl > 0)
    fprintf(debuglog, "function_ubee_r port %c\n", c);

 return (UINT16)(c);
}

//==============================================================================
// uBee512 write - Special Ubee512 Port function
//
// All commands pass a Z80 address to a structure in memory.  The structure
// contents is dependent on the command but all share the same overhead.  Data
// passed and returned in the structure is always in little endian format.
//
// IMPORTANT: This function does not currently address banking issues so it
// is important to realise that data may get lost if ANY bank swapping occurs
// after this function call.  To avoid problems do not switch banks or call
// any OS calls before the returned data is used.  Essentially the main area
// of concern is the 32K DRAM bank at 0x8000 that may also be at 0x0000.
// Normally writes to these banks will be copied from one to the other, but
// this does not occur here as we are acting on Z80 memory outside of the
// MZ80 context.  Also do not have the structure in video memory for similar
// reasons,  unlikely you would want to do that any way.
//
// It is important to verify that the application is running under the emulator
// and not a real Microbee before calling this function.  Use the uBee512 read
// port function to check first.
//
// The calling sequence is:
// port[n] = command number (0..255)
// port[n] = LSB of structure address (0 if not using a structure) 
// port[n] = MSB of structure address (0 if not using a structure)
//
// MEMBER	SIZE
// result	16b		result of the function call, 0=no error
// other        8b, 16b, 32b	other data passed/returned
// 
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void function_ubee_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
// uint16_t temp_u16_a;

// uint32_t temp_u32_a;
// uint32_t temp_u32_b;
// uint64_t temp_u64_a;
// uint64_t temp_u64_b; 
 
#if 0
 printf("function_ubee__w port 0x%02x = 0x%02x\n", port, data);
 return; 
#endif

 if (debuglvl > 0)
    fprintf(debuglog, "function_ubee_w port 0x%02x = 0x%02x\n", port, data);

 if (ubee_req)		// if we need to get the structure address
    {
     ubee_req--;    
     if (ubee_req == 1)
        ubee_addr = data;	// LSB of the address
     else   
        if (ubee_req == 0)
           {
            ubee_addr = ubee_addr | (data << 8);	// MSB of the address
            switch (ubee_command)
               {
                case 0x00 : break;
                case 0x10 : break;
                  default : break;          
               }
           } 
    }
 else
    {
     ubee_command = data;
     ubee_req = 2;
    }
}
