//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                            Memory mapper module                            *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Emulates the 128-512K memory bank switching and video active signal.
// Emulates the 32-56K RAM and 64K DRAM models.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.1.0 - 27 October 2007, S.J.Kay
// - boot_data had incorrect ordering of p256k, 256k, and p128k, 128k models,
//   moved the 256k models above the 128k models.
// - Added optional DRAM model ROMs for each model type.
// - Implement the modelx information structure.
//
// v2.0.0 - 17 October 2007, S.J.Kay
// - All FDD models can now load seperate boot ROMs and a fall-back ROM will be
//   loaded for some models if the default ROM is not found.
// - Added the Teleterm 256TC, PJB P256K, and 256K models.
// - APC model name changed to 56K for ROM loading.
//
// v1.4.0 - 26 September 2007, S.J.Kay
// - Changes for banked and non-banked models.  Models of 64K or less are
//   set up for VDU to fixed at address 0xF000.
// - Added alternive boot ROMs to be loaded depending on the model being
//   emulated,  for ROM machines this means a BASIC ROM.
// - Added memmap_basicrom() function for Premium ROM model basic upper/lower
//   half selection.
// - loadrom() function modified and is now global function memmap_loadrom()
// - Removed the non aplpha+ source build option.
// - Added changes to error reporting.
//
// v1.0.0 - 7 July 2007, S.J.Kay
// - added vdu_switchvideo_in() and vdu_switchvideo_out() function calls.
// - ROMs 1-3 required seperate handlers as there will be DRAM at 0xC000-0xDFFF
//   when ROM3 is selected.  This change allows the DRAM area to be written to.
//   A smaller dupwrite handler is also activated if the top and bottom banks
//   are the same and ROM3 is selected. (there is an 8K DRAM window).
// - According to the Technical manual ROM3 appears to be placed into wrong
//   memory location (0xC000) and should be placed into 0xE000.  Also ROM3 is
//   documented as being 8K and ROM1 and ROM2 are 16K.  Have made changes in
//   various locations for this and have modified the load_rom function and
//   optioned out the roundup function for now.
//
// - Converted to 512K DRAM model Microbee (as used for PJB 512K BIOS).  Now
//   uses array of pointers for banks.
//
// v0.0.0 - 5 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "memmap.h"
#include "ubee512.h"
#include "z80.h"
#include "vdu.h"
#include "mz80.h"

#include "macros.h"

static void switchin (UINT8 m);
static void switchout (int oldmode);

static void romxwrite (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s);
static void rom1write (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s);
static void rom2write (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s);
static void rom3write (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s);
static void dupwrite (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s);

static struct MemoryWriteByte z80_mem_w[MAXMEMHANDLERS] =
   { { -1, -1, NULL, NULL } };

static struct MemoryReadByte z80_mem_r[MAXMEMHANDLERS] =
   { { -1, -1, NULL, NULL } };

static char rom1[ROM1_SIZE],
            rom2[ROM2_SIZE],
            rom3[ROM3_SIZE];

static char bank00[BANKSIZE],
            bank01[BANKSIZE],
            bank02[BANKSIZE],
            bank03[BANKSIZE],
            bank04[BANKSIZE],
            bank05[BANKSIZE],
            bank06[BANKSIZE],
            bank07[BANKSIZE],
            bank08[BANKSIZE],
            bank09[BANKSIZE],
            bank10[BANKSIZE],
            bank11[BANKSIZE],
            bank12[BANKSIZE],
            bank13[BANKSIZE],
            bank14[BANKSIZE],
            bank15[BANKSIZE];

char const *bank_ptrs[] =
   {
    bank01, bank02, bank00, bank03,
    bank00, bank02, bank01, bank03,

    bank04, bank05, bank06, bank07,
    bank04, bank05, bank06, bank07,
    bank08, bank09, bank10, bank11,
    bank08, bank09, bank10, bank11,
    bank12, bank13, bank14, bank15,
    bank12, bank13, bank14, bank15
   };

char const *dram_bank_ptrs[] =
   {
    bank01, bank02, bank00, bank03,
    bank04, bank05, bank06, bank07,
    bank08, bank09, bank10, bank11,
    bank12, bank13, bank14, bank15
   };

int mem_mode;		// current memory map configuration (port 0x50)

static char basic[0x4000];	// storage area for Premium basic in ROM
static char romimage1[512];
static int x_upper;

extern int model;
extern model_t modelx;

extern char userhome_romspath[];

extern FILE *debuglog;
extern int debuglvl;

//==============================================================================
// Define all the possible ROM boot up options for disk and ROM based models.
// 
// The boot_data table must match the enumeration table order containing
// MOD_NAMES (in ubee512.h)
//==============================================================================
 static boot_t boot_data [MOD_TOTAL] =
    {
// MOD_256TC
     {
      {"256TC.ROM", "", "", ""},
      rom1,
      0,
      0x4000,
      "256TC_2.ROM",
      "256TC_3.ROM"
     },

// MOD_P512K
     {
      {"P512K.ROM", "rom1.bin", "", ""},
      rom1,
      0,
      0x2000,
      "P512K_2.ROM",
      "P512K_3.ROM"
     },

// MOD_512K
     {
      {"512K.ROM", "rom1.bin", "", ""},
      rom1,
      0,
      0x2000,
      "512K_2.ROM",
      "512K_3.ROM"
     },

// MOD_P256K
     {
      {"P256K.ROM", "rom1.bin", "", ""},
      rom1,
      0,
      0x2000,
      "P256K_2.ROM",
      "P256K_3.ROM"
     },

// MOD_256K
     {
      {"256K.ROM", "rom1.bin", "", ""},
      rom1,
      0,
      0x2000,
      "256K_2.ROM",
      "256K_3.ROM"
     },

// MOD_P128K
     {
      {"P128K.ROM", "rom1.bin", "", ""},
      rom1,
      0,
      0x2000,
      "P128K_2.ROM",
      "P128K_3.ROM"
     },

// MOD_128K    
     {
      {"128K.ROM", "rom1.bin", "", ""},
      rom1,
      0,
      0x2000,
      "128K_2.ROM",
      "128K_3.ROM"
     },

// MOD_P64K
     {
      {"P64K.ROM", "", "", ""},
      rom1,
      0,
      0x2000,
      "P64K_2.ROM",
      "P64K_3.ROM"
     },

// MOD_64K
     {
      {"64K.ROM", "", "", ""},
      rom1,
      0,
      0x2000,
      "64K_2.ROM",
      "64K_3.ROM"
     },

// MOD_56K
     {
      {"56K.ROM", "", "", ""},
      z80mem,
      0xE000,
      0x1000,
      "",
      ""
     },

// MOD_PPC85
     {
      {"PREM-A.ROM", "", "", ""},
      basic,
      0,
      0x4000,
      "",
      ""
     },

// MOD_PC85
     {
      {"PC85_BASIC.ROM", "", "", ""},
      z80mem,
      0x8000,
      0x4000,
      "",
      ""
     },

// MOD_PC
     {
      {"PC_BASIC.ROM", "PC_BASIC_A.ROM", "PC_BASIC_B.ROM", ""},
      z80mem,
      0x8000,
      0x4000,
      "",
      ""
     },

// MOD_IC
     {
      {"IC_BASIC.ROM", "IC_BASIC_A.ROM", "IC_BASIC_B.ROM", ""},
      z80mem,
      0x8000,
      0x4000,
      "",
      ""
     },

// MOD_2MHZ
     {
      {"2MHZ_BASIC.ROM", "2MHZ_BASIC_A.ROM", "2MHZ_BASIC_B.ROM", "2MHZ_BASIC_C.ROM", "2MHZ_BASIC_D.ROM", ""},
      z80mem,
      0x8000,
      0x4000,
      "",
      ""
     }
    };

//==============================================================================
// Load a boot or BASIC ROM image from file.
//
// Each model uses a different file name so that the user can link to the
// actual ROM file concerned.
//
//   pass: void
// return: int			0 if no errors, -1 if error
//==============================================================================
static int memmap_loadboot (void)
{
 int i;
 int loaded;
 int offset;
 int size;

 char romimage2[512];
 char romimage3[512]; 

 i = 0;
 loaded = 0;
 offset = boot_data[model].offset;
 size = boot_data[model].size;

 while (size > 0)
    {
     if (strcmp(boot_data[model].romimage[i], "") == 0)
        {
         console_output();
         printf("memmap_loadboot: Unable to continue ROM load operation (no ROM file specified).\n");
         return -1;
        }
     romimage1[0] = 0;
     strcat(romimage1, userhome_romspath);
     strcat(romimage1, boot_data[model].romimage[i]);
     loaded += memmap_loadrom(romimage1, boot_data[model].dest + offset + loaded, size);
     size -= loaded;
     i++;
#if 0
     printf("memmap_loadboot: file=%s size=%d loaded=%d\n", romimage1, size, loaded);
#endif     
    }    

 if (loaded != boot_data[model].size)
    {
     console_output();
     printf("memmap_loadboot: Unable to load %s (required to boot).\n", romimage1);
     return -1;
    }
 else
    if (! modelx.rom)
       printf("memmap_loadboot: boot ROM image loaded: %s\n", romimage1);     

// special treatment for Premium PC85 model
 if (model == MOD_PPC85)
    {
     memcpy(z80mem + 0x8000, basic, 0x2000);	// copy the 1st half of BASIC
     
     romimage1[0] = 0;
     strcat(romimage1, userhome_romspath);            
     strcat(romimage1, "PREM-B.ROM");
     if (memmap_loadrom(romimage1, z80mem + 0xA000, 0x2000) != 0x2000)
        {
         console_output();
         printf("memmap_loadboot: Unable to load PREM-B.ROM (required to boot).\n");
         return -1;
        }
    }

// Load the optional ROMs for the DRAM models
 if (modelx.ram >= 64)
    {
     snprintf(romimage2, sizeof(romimage2), "%s%s", userhome_romspath, boot_data[model].romimage2);
     snprintf(romimage3, sizeof(romimage3), "%s%s", userhome_romspath, boot_data[model].romimage3);     

     if (memmap_loadrom(romimage2, rom2, ROM2_SIZE))
        printf("memmap_loadboot: optional ROM 2 image loaded: %s\n", romimage2);
     if (memmap_loadrom(romimage3, rom3, ROM3_SIZE))
        printf("memmap_loadboot: optional ROM 3 image loaded: %s\n", romimage3);
    } 
 
 return 0;
}

//==============================================================================
// Initialise the memory map.
//
// Bank memory handling is used for models with more than 64K of RAM/DRAM.
// Models with 64K or less do not have bank switching and the VDU is fixed
// at 0xF000.
//
//   pass: void
// return: int			0 if no errors, -1 if error
//==============================================================================
int memmap_init (void)
{
 struct mz80context z80;

 int rindex = 0;
 int windex = 0;

 mz80GetContext(&z80);
 z80.z80MemRead = z80_mem_r;
 z80.z80MemWrite = z80_mem_w;
 mz80SetContext(&z80);

// For debugging purposes only, fills each DRAM bank with numbers.
#ifdef use_dramdebug
     int i;
     for (i = 0; i < 16; i++)
        memset(dram_bank_ptrs[i], i, BANKSIZE);
#endif

 if (memmap_loadboot() == -1)
    return -1;
 
 if (modelx.ram >= 64)
    {
// initialise the mamory map to default
     mem_mode = -1; 
     memmap_mode(0x50, 0, NULL);
    }
 else
// Initialise for RAM models (< 64K)
    {
// VDU write handler
     z80_mem_w[windex].memoryCall = vdu_vidmem_w;
     z80_mem_w[windex].pUserArea  = NULL;
     z80_mem_w[windex].lowAddr = 0xF000;
     z80_mem_w[windex].highAddr = 0xFFFF;
     windex++;

// ROM write handler
     if (modelx.rom)
        {
         z80_mem_w[windex].lowAddr = 0x8000;
         z80_mem_w[windex].highAddr = 0xEFFF;
        } 
     
     if (model == MOD_56K)
        {
         z80_mem_w[windex].lowAddr = 0xE000;
         z80_mem_w[windex].highAddr = 0xEFFF;
        } 

     z80_mem_w[windex].memoryCall = romxwrite;
     z80_mem_w[windex].pUserArea = NULL;
     windex++;

// Terminate write structure
     z80_mem_w[windex].lowAddr = -1;
     z80_mem_w[windex].highAddr = -1;
     z80_mem_w[windex].memoryCall = NULL;
     z80_mem_w[windex].pUserArea = NULL;

// VDU read handler
     z80_mem_r[rindex].memoryCall = vdu_vidmem_r;
     z80_mem_r[rindex].pUserArea = NULL;
     z80_mem_r[rindex].lowAddr = 0xF000;
     z80_mem_r[rindex].highAddr = 0xFFFF;
     rindex++;

// Terminate read structure
     z80_mem_r[rindex].lowAddr = -1;
     z80_mem_r[rindex].highAddr = -1;
     z80_mem_r[rindex].memoryCall = NULL;
     z80_mem_r[rindex].pUserArea = NULL;

// initialise the memory map to default, this is used in other modules and
// must be set to VDU at 0xF000, VDU enabled, etc

     mem_mode = 0;
     vdu_switchvideo_in();
    }
 
 return 0;
}

//==============================================================================
// De-init the memory map (N/A)
//
//   pass: void
// return: int			0
//==============================================================================
int memmap_deinit (void)
{
 return 0;
}

//==============================================================================
// Set memory map to the reset condition
//
//   pass: void
// return: int			0
//==============================================================================
int memmap_reset (void)
{
 if (modelx.ram >= 64)
    memmap_mode(0x50, 0, NULL);
    
 if (modelx.rom)
    {
     x_upper = 1;
     memmap_basicrom(0);
     vdu_switchvideo_out();     
     vdu_switchvideo_in();
    } 

 return 0;
}

//==============================================================================
// Load a ROM image from file.
//
// The messages reported in this function are not considered as fatal so are
// not forced to the console.
//
//   pass: char *name  		pointer to file name
//         char *dest  		pointer to destination
//         int size 		expected size of the ROM to be loaded
// return: int			number of bytes loaded
//==============================================================================
int memmap_loadrom (char *name, char *dest, int size)
{
 FILE *romfp;
 long len;

 if (dest == NULL)
    return 0;

 memset(dest, 0, size);

 romfp = fopen(name, "rb");
 if (romfp == NULL)
    return 0;

 fseek(romfp, 0, SEEK_END);  		// get size of ROM
 len = ftell(romfp);
 fseek(romfp, 0, SEEK_SET);		// back to start

 if (len != 0)
    {
     if (len > size)
        {
         printf("memmap_loadrom: ROM size is too large (continuing): %s\n", name);
         return 0;
        }
     else
        if (len < size)
           size = len;
     len = fread(dest, 1, len, romfp);
     fclose(romfp);
    } 

 if (len == 0)
    printf("memmap_loadrom: Failed to load ROM (continuing): %s\n", name);

 return len;
}

//==============================================================================
// Set memory map mode - Port 0x50 write handler
//
//   pass: UINT16 port
//	   UINT8 data
//         struct z80PortWrite *port_struct
// return: void
//==============================================================================
void memmap_mode (UINT16 port, UINT8 data, struct z80PortWrite *port_struct)
{
 if (modelx.ram < 64)
    return;

 if (debuglvl > 0)
    fprintf(debuglog, "memmap_mode set = 0x%02x\n", data);

#ifdef use_debug_memmap_mode
   printf("memmap_mode set = 0x%02x\n", data);
#endif
 
 if (mem_mode != -1)
    switchout(mem_mode);

 switchin(data); 
}

//==============================================================================
// Switch in the lower or upper half of the Premium ROM basic if emulating a
// Premium PC85 ROM machine.
//
//   pass: int upper			1 for upper, else lower
// return: void
//==============================================================================
void memmap_basicrom (int upper)
{
 if ((model == MOD_PPC85) && (upper != x_upper))
    {
     if (upper)
        memcpy(z80mem + BASIC1_ADDR, basic + 0x2000, 0x2000);
     else   
        memcpy(z80mem + BASIC1_ADDR, basic, 0x2000);
     x_upper = upper;
    }
}        

//==============================================================================
// Switch in DRAM banks, Video memory locations, ROMs 1,2 and 3, etc
//
//   pass: int m
// return: void
//==============================================================================
static void switchin (UINT8 m)
{
 int rindex = 0;
 int windex = 0;
 int banksel;
 char *bank;

 switch (modelx.ram)
    {
     case 512 : mem_mode = m & B8(11111111);
                banksel = ((mem_mode & B8(11000000)) >> 3) | (mem_mode & B8(00000111));
                break;
     case 256 : if (model == MOD_256TC)
                   {
                    mem_mode = m & B8(00111111);
                    banksel = ((mem_mode & B8(00100000)) >> 2) | (mem_mode & B8(00000111));
                   }
                else
                   {   
                    mem_mode = m & B8(01111111);
                    banksel = ((mem_mode & B8(01000000)) >> 3) | (mem_mode & B8(00000111));
                   } 
                break;
     case 128 : mem_mode = m & B8(00111111);
                banksel = mem_mode & B8(00000111);
                break;
     case  64 : mem_mode = m & B8(00111101);
                banksel = mem_mode & B8(00000101);
                break;
      default : banksel = 0;
    }
    
 bank = (char *)bank_ptrs[banksel];

// Copy DRAM bank to memory at 0x0000-0x7FFF
 memcpy(z80mem, bank, BANKSIZE);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// insert video memory handler if video RAM is enabled.
//
// Must appear early in the list of handlers so it overrides anything that
// overlaps it.  The handler that gets called is the first one with matching
// range, as a result of the mz80 implementation.  This might change in the
// future!
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 if (! (mem_mode & BANK_VRAM))	// if video RAM enabled
   {
    z80_mem_r[rindex].memoryCall = vdu_vidmem_r;
    z80_mem_r[rindex].pUserArea  = NULL;

    z80_mem_w[windex].memoryCall = vdu_vidmem_w;
    z80_mem_w[windex].pUserArea  = NULL;

    if (mem_mode & BANK_VADD)   // if video RAM at 8000-8FFF
       {
        z80_mem_r[rindex].lowAddr =  0x8000;
        z80_mem_r[rindex].highAddr = 0x8FFF;        

        z80_mem_w[windex].lowAddr =  0x8000;
        z80_mem_w[windex].highAddr = 0x8FFF;
       }
    else
       {
        z80_mem_r[rindex].lowAddr =  0xF000;
        z80_mem_r[rindex].highAddr = 0xFFFF;

        z80_mem_w[windex].lowAddr =  0xF000;
        z80_mem_w[windex].highAddr = 0xFFFF;
       }
    rindex++;
    windex++;
   }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Test if any ROMs are in and set up handlers accordingly.
//
// - if no ROMs in then copy the complete memory bank BANKUBLK to upper
//   memmory.
// - if ROM3 is in then there is an 8K window of DRAM at 0xC000-DFFF and this
//   copyed to from BANKUBLK.
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 if (mem_mode & BANK_ROMS)     		// if ROMs disabled
    {
     memcpy(z80mem+BANKSIZE, BANKUBLK, BANKSIZE);  // copy bank

     if ((mem_mode & BANK_DRAM) == BANK_HIGH)  // if high bank
        {
// this bank is mapped into upper and lower memory , so we need to echo any
// writes to one half to the other.
         z80_mem_w[windex].lowAddr = 0x0000;
         z80_mem_w[windex].highAddr = 0xFFFF;
         z80_mem_w[windex].memoryCall = dupwrite;
         z80_mem_w[windex].pUserArea = NULL;
         windex++;
        }
    }
 else
    {
     z80_mem_w[windex].lowAddr = 0x8000;
     z80_mem_w[windex].highAddr = 0xBFFF;
     z80_mem_w[windex].memoryCall = rom1write;
     z80_mem_w[windex].pUserArea = NULL;
     windex++;    
     memcpy(z80mem+0x8000, rom1, ROM1_SIZE);
     
     if (mem_mode & BANK_ROMX)   // if ROM3 selected
        {
         memcpy(z80mem+0xC000, BANKUBLK+0x4000, 0x2000);  // copy 8K only        

         if ((mem_mode & BANK_DRAM) == BANK_HIGH)  // if high bank
            {
// this bank is mapped into upper and lower memory , so we need to echo any
// writes to one half to the other.
             z80_mem_w[windex].lowAddr = 0x04000;
             z80_mem_w[windex].highAddr = 0x5FFF;
             z80_mem_w[windex].memoryCall = dupwrite;
             z80_mem_w[windex].pUserArea = NULL;
             windex++;
            }
         z80_mem_w[windex].lowAddr = 0xC000; // need to handle some DRAM too
         z80_mem_w[windex].highAddr = 0xFFFF;
         z80_mem_w[windex].memoryCall = rom3write;
         z80_mem_w[windex].pUserArea = NULL;
         windex++;     
         memcpy(z80mem+0xE000, rom3, ROM3_SIZE);
        } 
     else
        {
         z80_mem_w[windex].lowAddr = 0xC000;
         z80_mem_w[windex].highAddr = 0xFFFF;
         z80_mem_w[windex].memoryCall = rom2write;
         z80_mem_w[windex].pUserArea = NULL;
         windex++;        
         memcpy(z80mem+0xC000, rom2, ROM2_SIZE);
        } 
    }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Video memory.
//
// Copy video memory into the main memory if video is enabled, done last so it
// overwrites anything already there.
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 if (! (mem_mode & BANK_VRAM))	// if video RAM enabled
    vdu_switchvideo_in();

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// terminate the memory handlers
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 z80_mem_r[rindex].lowAddr = -1;
 z80_mem_r[rindex].highAddr = -1;
 z80_mem_r[rindex].memoryCall = NULL;
 z80_mem_r[rindex].pUserArea = NULL;

 z80_mem_w[windex].lowAddr = -1;
 z80_mem_w[windex].highAddr = -1;
 z80_mem_w[windex].memoryCall = NULL;
 z80_mem_w[windex].pUserArea = NULL;
}

//==============================================================================
// Switch out DRAM banks, Video memory locations, ROMs 1,2 and 3, etc
//
//   pass: int oldmode
// return: void
//==============================================================================
static void switchout (int oldmode)
{
 int banksel;
 char *bank;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Copy DRAM memory at 0000-7FFF to bank
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 switch (modelx.ram)
    {
     case 512 : banksel = ((oldmode & B8(11000000)) >> 3) | (oldmode & B8(00000111));
                break;
     case 256 : if (model == MOD_256TC)
                   banksel = ((oldmode & B8(00100000)) >> 2) | (oldmode & B8(00000111));
                else
                   banksel = ((oldmode & B8(01000000)) >> 3) | (oldmode & B8(00000111));
                break;
     case 128 : banksel = oldmode & B8(00000111);
                break;
     case  64 : banksel = oldmode & B8(00000101);
                break;                
      default : banksel = 0;
    }
 bank = (char *)bank_ptrs[banksel];
 
// Copy DRAM memory at 0000-7FFF to bank
 memcpy(bank, z80mem, BANKSIZE);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Switch video RAM out if enabled.
//
// NOTE: use of vb & B8(00000011) for screen, attribute and colour RAM as
// these are at most 8K each (4 x 2K banks).
//
// The vb value used in the switch in function will be used in the switch out
// so that the correct video RAM banks can be switched.
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 if (!(oldmode & BANK_VRAM))  // if video RAM enabled
    {
     vdu_switchvideo_out();

     if (oldmode & BANK_VADD) // if video at 8000-8FFF
        {
         if (oldmode & BANK_ROMS)	// if ROMs disabled
            {
             if ((oldmode & BANK_DRAM) != BANK_HIGH)  // if NOT high bank
                memcpy(BANKUBLK+0x1000, z80mem+0x9000, 0x7000);
            }
         else
            {
             if (((oldmode & BANK_DRAM) != BANK_HIGH) && (oldmode & BANK_ROMX))   // if ROM3 selected             
                memcpy(BANKUBLK+0x4000, z80mem+0xC000, 0x2000);                   
            }    
        }
     else
        {
         if (oldmode & BANK_ROMS)    	// if ROMs disabled
            {
             if ((oldmode & BANK_DRAM) != BANK_HIGH)  // if NOT high bank
                memcpy(BANKUBLK, z80mem+0x8000, 0x7000);
            }
         else
            { 
             if (((oldmode & BANK_DRAM) != BANK_HIGH) && (oldmode & BANK_ROMX))   // if ROM3 selected             
                memcpy(BANKUBLK+0x4000, z80mem+0xC000, 0x2000);                   
            }       
        }
    }
 else

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// no video RAM, or ROMS enabled, copy current memory map 8000-FFFF back to it's
// bank location.
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    if (oldmode & BANK_ROMS) 		// if ROMs are disabled
       if ((oldmode & BANK_DRAM) != BANK_HIGH)  // if NOT high bank
// copy memory map 8000-FFFF to bank
          memcpy(BANKUBLK, z80mem+BANKSIZE, BANKSIZE);         
}

//==============================================================================
// ROM x write (0x8000-0xEFFF)
//
// This handler is installed for models with 64K or less RAM/DRAM
// map. No writing to any ROM areas will occur.
//
//   pass: UINT32 addr
//	   UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
static void romxwrite (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
}

//==============================================================================
// ROM 1 write (0x8000-0xBFFF)
//
// This handler is installed when ROM1 (port 0x50 bit 2=0) are in the memory
// map. No writing to any ROM areas will occur.
//
//   pass: UINT32 addr
//	   UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
static void rom1write (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
}

//==============================================================================
// ROM 2 write (0xC000-0FFFF)
//
// This handler is installed when ROM2 (port 0x50 bit 2=0) and the BANK_ROMX
// bit is clear in mem_mode.  No writing to any ROM areas will occur.
//
//   pass: UINT32 addr
//	   UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
static void rom2write (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
}

//==============================================================================
// ROM 3 write (0xE000-0FFFF)
//
// This handler is installed when ROM3 (port 0x50 bit 2=0) and the BANK_ROMX
// bit is set in mem_mode.  No writing to any ROM areas will occur.
//
// If the memory access address is DRAM (0xC000-0xDFFF) then a write will
// occur to DRAM memory.
//
// If BANKUBLK is in context at 0x0000 then any DRAM write (0xC000-0xDFFF)
// must also be written to DRAM (0x4000-0x5FFF)
//
//   pass: UINT32 addr
//	   UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
static void rom3write (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
 if ((addr > 0xBFFF) && (addr < 0xE000))
    {
     z80mem[addr] = data;
     if ((mem_mode & BANK_DRAM) == BANK_HIGH)  // if high bank
        z80mem[addr & 0x7FFF] = data;
    }
}

//==============================================================================
// duplicates writes to one half of DRAM to the other half.  This handler is
// installed when DRAM bank 0, block 1 is in the lower half of memory as it is
// also in the upper half of memory.  Depending on the selection of ROM3
// the address range for this handler activation is different.
//
//   pass: UINT32 addr
//         UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
static void dupwrite (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
 z80mem[addr & 0x7FFF] = data;
 z80mem[addr | 0x8000] = data;
}
