//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                        Real Time Clock (RTC) module                        *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Emulates the MC146818 RTC circuit.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.1.0 - 20 October 2007, S.J.Kay
// - Added alarm flag and interrupt set to rtc_poll() function.
// - Changes to return the AM/PM hours and alarm values correctly.
// - Changes to save RTC image with data that a real MC146818 would contain.
// - Implement the modelx information structure.
//
// v2.0.0 - 15 October 2007, S.J.Kay
// - Created a new file and implement the RTC emulation.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

//#ifdef MINGW
//#include <windows.h>
//#else
//#endif

#include "rtc.h"
#include "ubee512.h"
#include "mz80.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
static UINT8 addr;

static rtc_u rtc;

static char name[512];

extern uint64_t z80_cycles;

extern int model;
extern model_t modelx;

extern char userhome_rtcpath[];
extern char *model_args[];

extern FILE *debuglog;
extern int debuglvl;

//==============================================================================
// Set clock times from host. 
//
// struct tm
//
// This structure is used to hold the values representing a calendar time.
// It contains the following members, with the meanings as shown.
//
// int tm_sec      /* seconds after minute [0-61] (61 allows for 2 leap-seconds)*/
// int tm_min      /* minutes after hour [0-59] */
// int tm_hour     /* hours after midnight [0-23] */
// int tm_mday     /* day of the month [1-31] */
// int tm_mon      /* month of year [0-11] */
// int tm_year     /* current year-1900 */
// int tm_wday     /* days since Sunday [0-6] */
// int tm_yday     /* days since January 1st [0-365] */
// int tm_isdst    /* daylight savings indicator */
//
//   pass: void
// return: void
//==============================================================================
static void rtc_setclockfromhost (void)
{
 typedef struct tm tm_t;

 time_t result;
 tm_t resultp;
 
 time(&result); 
#ifdef MINGW    
 memcpy(&resultp, localtime(&result), sizeof(resultp));
#else
 localtime_r(&result, &resultp);
#endif 
 if (resultp.tm_sec > 59)		// 2 second leap-seconds ignore
    resultp.tm_sec = 59;
 rtc.member.seconds = resultp.tm_sec;
 rtc.member.minutes = resultp.tm_min;                  
 rtc.member.hours = resultp.tm_hour;
 rtc.member.mday = resultp.tm_mday;
 rtc.member.month = resultp.tm_mon + 1;
 rtc.member.year = resultp.tm_year - 100;     
 rtc.member.wday = resultp.tm_wday + 1;          
}

//==============================================================================
// Convert time data value to format specified by RTC_B_DM
//
//   pass: int time			time in binary or BDC format
// return: int				time in format specified by RTC_B_DM
//==============================================================================
static inline int rtc_time_convert (int time)
{
 if ((rtc.member.reg_b & RTC_B_DM) == 0)     // if BCD format required
    time = ((time / 10) << 4) | (time % 10);
 return time;
}     

//==============================================================================
// Convert time data value to binary format if RTC_B_DM is 0
//
//   pass: int time			time in binary or BCD format
// return: int				time in binary format
//==============================================================================
static inline int rtc_time_convert_bcdtobin (int time)
{
 if ((rtc.member.reg_b & RTC_B_DM) == 0)    // if in BCD format
    time = ((time >> 4) * 10) + (time & 0x0f);
 return time;
}     

//==============================================================================
// Convert binary 24 hour time format to the format as specified by RTC_B_2412
// and RTC_B_DM bits.
//
//   pass: int hours			hours in AM/PM or 24 hour format
// return: int				hours in required format
//==============================================================================
static inline int rtc_hours_convert (int hours)
{
 if ((rtc.member.reg_b & RTC_B_2412) == 0)
    {
     if (hours == 0)       		// if 12 AM midnight
        return rtc_time_convert(12);
     else
        if (hours > 12)
           return rtc_time_convert(hours % 12) | B8(10000000);
        else
           return rtc_time_convert(hours);
    }
 else
    return rtc_time_convert(hours);    
}

//==============================================================================
// Convert hours value to 24 hour binary format if RTC_B_2412 is 0
//
//   pass: int hours			hours in AM/PM or 24 hour format
// return: int				hours in 24 hour format
//==============================================================================
static inline int rtc_hours_convert_12to24 (int hours)
{
 int pm;
 
 pm = hours & B8(10000000);
 hours = rtc_time_convert_bcdtobin(hours & B8(01111111));

 if ((rtc.member.reg_b & RTC_B_2412) == 0)  // if AM/PM mode
    {
     if ((hours == 12) && (pm))		// if 12 PM midday
        return 12;			// 1200
     if (pm)
        return hours + 12;		// 1300 - 2300
     if (hours == 12)		        // if 12 AM midnight
        return 0;			// 0000
     return hours;			// 0100 - 1100
    }

 return hours;
}

//==============================================================================
// RTC Initialise.
//
// Load the RTC file for the model being emulated if one exists.
//
//   pass: void
// return: int				0 if success, -1 if error
//==============================================================================
int rtc_init (void)
{
 FILE *fp;
 long len;

 if (modelx.rtc)
    {
     snprintf(name, sizeof(name), "%s%s.rtc", userhome_rtcpath, model_args[model]);
     fp = fopen(name, "rb");

     if (fp != NULL)
        {
         fseek(fp, 0, SEEK_END);  	        // get size of the RTC file
         len = ftell(fp);
         fseek(fp, 0, SEEK_SET);		// back to start

         if (len != sizeof(rtc))
            {
             console_output();
             printf("rtc_init: RTC file is incorrect size: %s\n", name);
             fclose(fp);
             return -1;
            }
         else
            fread(&rtc, 1, sizeof(rtc), fp);
         fclose(fp);
        }
     else
        memset(&rtc, 0, sizeof(rtc));
        
     rtc.member.reg_d = RTC_D_VRT;
     addr = 0;
     rtc_setclockfromhost();
    }
 return 0;
}

//==============================================================================
// RTC de-initialise.
//
// Before the RTC memory is saved some values need to be converted to the
// correct format as would be found inside the RTC IC.
//
//   pass: void
// return: int				0
//==============================================================================
int rtc_deinit (void)
{
 FILE *fp;

 if (modelx.rtc)
    {
     rtc.member.seconds = rtc_time_convert(rtc.member.seconds);
     rtc.member.minutes = rtc_time_convert(rtc.member.minutes);
     rtc.member.hours = rtc_hours_convert(rtc.member.hours);
     rtc.member.mday = rtc_time_convert(rtc.member.mday);
     rtc.member.month = rtc_time_convert(rtc.member.month); 
     rtc.member.year = rtc_time_convert(rtc.member.year);
     rtc.member.wday = rtc_time_convert(rtc.member.wday);

     snprintf(name, sizeof(name), "%s%s.rtc", userhome_rtcpath, model_args[model]);
     fp = fopen(name, "wb");

     if (fp != NULL)
        {
         fwrite(&rtc, 1, sizeof(rtc), fp);
         fclose(fp);
         return 0;
        }
     else
        {
         console_output();
         printf("rtc_deinit: Unable to create RTC file: %s\n", name);
         return -1;
        }
    }
 return 0;
}

//==============================================================================
// RTC reset.
//
//   pass: void
// return: int				0
//==============================================================================
int rtc_reset (void)
{
 if (modelx.rtc)
    {
     rtc.member.reg_b &= (0xff ^ (RTC_B_PIE | RTC_B_AIE | RTC_B_UIE | RTC_B_SQWE));
     rtc.member.reg_c = 0;
    } 
 return 0;
}

//==============================================================================
// RTC read register data - Port function
//
// The alarm values do not require conversion as these are set directly by the
// application program.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16 register 		register data
//==============================================================================
UINT16 rtc_r (UINT16 port, struct z80PortRead *port_s)
{
 uint8_t data;
 
 if (modelx.rtc)
    {
#if 0
     if ((standard != 0) && (port > 0x07))
        return 0;
#endif
     switch (port & 0x0f)
        {
         case 0x04 : if (debuglvl > 0)
                        fprintf(debuglog, "rtc_r: port=0x%02X data=0x%02X\n", port, addr);
//                     printf("rtc_r: port=0x%02X data=0x%02X\n", port, addr);
                     return addr;
         case 0x06 : if (debuglvl > 0)
                        fprintf(debuglog, "rtc_r: port=0x%02X", port);
//                     printf("rtc_r: port=0x%02X", port);
                     return 0;
         case 0x07 : data = rtc.ram[addr];
                     if (addr < reg_a)
                        {
                         if (addr == hours)
                            data = rtc_hours_convert(data);
                         else
                            if ((addr != hours_alarm) && (addr != minutes_alarm) && (addr != seconds_alarm))
                               data = rtc_time_convert(data);
                        }
                     else
                        if (addr == reg_c)
                           rtc.member.reg_c = 0;	// clear all flag bits
                           
                     if (debuglvl > 0)
                        fprintf(debuglog, "rtc_r: port=0x%02X data=0x%02X\n", port, data);
//                     printf("rtc_r: port=0x%02X data=0x%02X\n", port, data);
                     return data;
        }            
    }
 return 0;
}

//==============================================================================
// RTC write register data - Port function
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void rtc_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.rtc)
    {
#if 0
     if ((standard != 0) && (port > 0x07))
        return 0;
#endif
     switch (port & 0x0f)
        {
         case 0x04 : addr = (data & 0x3F);
                     break;
         case 0x06 : if (addr < reg_a)
                        rtc.ram[addr] = data;
                     else
                        if (addr == reg_b)
                           {
                            if ((data & RTC_B_SET) && ((rtc.ram[addr] & RTC_B_SET) == 0))
                               data &= (0xff ^ RTC_B_UIE);
                            rtc.ram[addr] = data;
                           }
                        else
                           if (addr > reg_d)
                              rtc.ram[addr] = data;
                     break;
         case 0x07 : break;
        }             
     if (debuglvl > 0)
        fprintf(debuglog, "rtc_w: port=0x%02X data=0x%02X\n", port, data);
//     printf("rtc_w: port=0x%02X data=0x%02X\n", port, data);
    }
}

//==============================================================================
// RTC poll
//
// Read the host system time and check for any RTC flags that may need setting
// and any interrupts to be generated.
//
//   pass: void
// return: int				RTC interrupt status
//==============================================================================
int rtc_poll (void)
{
 static int secs1 = -1;
 static int wait_newtime = 0;
 int secs;

 if (modelx.rtc)
    {
     rtc_setclockfromhost();

// Check and set alarm flag and interrupts if time matches
     if (
         (rtc_time_convert_bcdtobin(rtc.member.seconds_alarm) == rtc.member.seconds) &&
         (rtc_time_convert_bcdtobin(rtc.member.minutes_alarm) == rtc.member.minutes) &&
         (rtc_hours_convert_12to24(rtc.member.hours_alarm) == rtc.member.hours)
        ) 
        {
         if (! wait_newtime)
            {
             wait_newtime = 1;
             rtc.member.reg_c |= RTC_C_AF;      // set alarm flag
             if (rtc.member.reg_b & RTC_B_AIE)
                rtc.member.reg_c |= RTC_C_IRQF;	// set IRQF flag
            }
        }    
     else
        wait_newtime = 0;

// Check and set Periodic and Update ended flags and interrupts
     secs = (z80_cycles / 3375000);
     if (secs != secs1) 
        {
         secs1 = secs;

         rtc.member.reg_c |= RTC_C_PF;		// set periodic flag
         rtc.member.reg_c |= RTC_C_UF;		// set update ended flag

         if ((rtc.member.reg_b & rtc.member.reg_c) & (RTC_B_PIE | RTC_B_UIE))
            rtc.member.reg_c |= RTC_C_IRQF;	// set IRQF flag
        }

     if (rtc.member.reg_c & RTC_C_IRQF)
        return B8(10000000);
    }
    
 return 0;
}
