//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                              Serial module                                 *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// This module is used to emulate the standard serial port.  It will require
// that the emulator is configured to match the settings used in a Microbee
// application.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.4.0 - 26 September 2007, S.J.Kay
// - Added changes to error reporting.
//
// v1.3.0 - 1 September 2007, S.J.Kay
// - Created a new file and implement the serial emulation at the top level.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#if MINGW
#include <windows.h>
#include <process.h>
#include <time.h>
#include <ctype.h>
#else
#endif

#include "ubee512.h"
#include "pio.h"
#include "async.h"
#include "serial.h"
#include "mz80.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
int serialrx = -1;

static int serial_databits;
static int serial_stopbits;

int serial_byte_rx=-1;
static uint64_t cycles_before_rx;
static int serial_divval_rx;

static uint64_t cycles_before_tx;
static uint8_t serial_before_tx=SERIAL_STOPBIT_TX;
static uint8_t serial_byte_tx;
static int serial_bitcount_tx;
static int serial_status_tx;
static int serial_divval_tx;

deschand_t coms1 = -1;		// file descriptor/handle
extern char *coms1_dev;
extern char coms1dev[];

extern pio_t pio_b;

extern uint64_t z80_cycles;

//==============================================================================
// Serial Initialise. 
//
// Open the serial input file and read in the WAV file header.
//
//   pass: void
// return: int				0 if success, -1 if error
//==============================================================================
int serial_init (void)
{
 if (coms1_dev != NULL)
    {
     coms1 = async_open(coms1dev);
     serialrx = -1;
     if (coms1 == (deschand_t)-1)
        {
         console_output();
         printf("serial_init: Failed to open serial device: %s\n", coms1dev);
         return -1;
        } 
    }
 return 0;
}

//==============================================================================
// Serial de-initialise.
//
// For serial out the WAV file header is filled in and the file closed.
// For serial in the file is closed.
//
//   pass: void
// return: int				0
//==============================================================================
int serial_deinit (void)
{
 if ((coms1_dev != NULL) && (coms1 != (deschand_t)-1))
    return async_close(coms1);
 return 0;
}

//==============================================================================
// Serial reset.
//
//   pass: void
// return: int				0
//==============================================================================
int serial_reset (void)
{
 if ((coms1_dev != NULL) && (coms1 != (deschand_t)-1))
    serialrx = -1;    
 return 0;
}

//==============================================================================
// Serial read poll.
//
// This can be called from the PIO module if serial receive interrupts are
// currently enabled.  If not then the serial_r() function will call this
// function.
//
//   pass: void
// return: int				serial character or -1 if none ready
//==============================================================================
int serial_readpoll (void)
{
 if ((coms1_dev != NULL) && (coms1 != (deschand_t)-1) && (serial_byte_rx == -1))
    {
     serialrx = async_read(coms1);
     return serialrx;
    } 
 return -1;   
}
 
//==============================================================================
// Serial read. Only call this if the serial is open for input.
//
// Once a character has started to be churned out, the serial_byte_rx
// variable will no longer be -1 until the churning process has completed. 
// It should be kept in mind that this process can be started by any read of
// this port for other unrelated reasons and this is fine and it all works
// as if it was a real Microbee but the serial_byte_rx varible may be left
// with a value that is not -1 and so an interrupt process may not be able
// to get started.  The PIO polling process should call this function if the
// serial_byte_rx variable is not -1 and PIO bort B interrupts are in
// affect. This will churn out the current serial byte and eventually
// becomes equal to -1.
//
// This all works because the data bits are based on the current Z80 cycle
// count and not the number of calls made to this function and so is not
// robbing any other calls of any data.
//
// The Microbee port does not invert data for RX.
// 
//   pass: void
// return: int				serial input bit value
//==============================================================================
int serial_r (void)
{
 int count;
 
 static uint64_t cycles_now_rx;
 static uint64_t cycles_elapsed_rx;

// if we got here via an interrupt we must adjust for lost time on the start
// bit as the interrupt serice routine consumes some cycles.  This is
// especially important at the faster baud rates.  As interrupts are
// serviced when re-entering or exiting the mz80exec() (in ubee512.c) the
// cycles that were recorded is actually what we want without adding in any
// elapsed cycle counts.

 if ((pio_b.ienableff != 0) && (serial_byte_rx == -1))
    {
     cycles_now_rx = z80_cycles;
     cycles_elapsed_rx = cycles_now_rx - cycles_before_rx;
    }
 else
    {   
     cycles_now_rx = z80_cycles + mz80GetElapsedTicks(0); // use 0 !
     cycles_elapsed_rx = cycles_now_rx - cycles_before_rx;
    } 
 
// calculate the number of serial bit times we have so far
 count = cycles_elapsed_rx / serial_divval_rx;

 if (serial_byte_rx == -1) 
    {
     if (pio_b.ienableff == 0)
        serial_readpoll();

     serial_byte_rx = serialrx;

     if (serial_byte_rx != -1)          // if we have some data to work with              
        {
         cycles_before_rx = cycles_now_rx;
         serial_byte_rx ^= 0xff;
         serial_byte_rx &= ((1 << serial_databits) - 1); // keep only n data bits
         serial_byte_rx = (serial_byte_rx << 1) | 0x01;  // add in the start bit
         count = 0;		// start from the beginning of the character
        }
    }
    
// return stop bits if no data is ready     
 if (serial_byte_rx == -1)
    return B8(00001000);	// stop bit, and CTS always true for now

// return start and data bits depending on count value
 if (count < (serial_databits + 1))
    return (((serial_byte_rx >> count) & 0x01) << 4) | B8(00001000); // get count pos bit
  
// end of the character, reset for a new character
 serial_byte_rx = -1;
 return B8(00001000);	// stop bit, and CTS always true for now
} 

//==============================================================================
// Serial write.  Only call this if the serial output is known to be open.
//
// The Microbee port inverts data for TX in hardware after the PIO.
//
//   pass: int data
// return: void
//==============================================================================
void serial_w (uint8_t data)
{
 int count;
 uint8_t serial_bit;
 
 static uint8_t serial_now_tx;
 
 static uint64_t cycles_now_tx;
 static uint64_t cycles_elapsed_tx;

// mask out the serial output bit
 serial_now_tx = (data & B8(00100000)) >> 5;
 
 cycles_now_tx = z80_cycles + mz80GetElapsedTicks(0); // use 0 !
 cycles_elapsed_tx = cycles_now_tx - cycles_before_tx;    

// determine the number of bit times we have so far
 count = cycles_elapsed_tx / serial_divval_tx;

// if we are after a start bit for a new byte
 if (serial_status_tx == 0)      // if a start bit is expected
    {
     cycles_before_tx = cycles_now_tx;
     if (serial_now_tx == SERIAL_STARTBIT_TX)
        {
         serial_status_tx++;     // want start bit delay
         serial_byte_tx = 0;
         serial_bitcount_tx = 0;
         serial_before_tx = serial_now_tx;
        }
     return;   
    }     

// must wait for at least 1 bit time to occur
 if (! count)
    return;

// if we are after the start bit time
 if (serial_status_tx == 1)      // if a start bit time is needed
    {
     cycles_before_tx = cycles_now_tx;
     serial_status_tx++;     // want data bits

     if (serial_now_tx != serial_before_tx)
        serial_before_tx = serial_now_tx;
     return;   
    }     

 if (serial_status_tx == 2)      // if data building in progress
    {
     cycles_before_tx = cycles_now_tx;     

     serial_bit = serial_before_tx;
     
     if (serial_now_tx != serial_before_tx)
         serial_before_tx = serial_now_tx;

     while ((count) && (serial_bitcount_tx != serial_databits))
        {
         serial_byte_tx |= (serial_bit << serial_bitcount_tx); 
         serial_bitcount_tx++; 
         count--;
        }        
    
     if (serial_bitcount_tx >= serial_databits)     
        {
         serial_status_tx = 0;     // want start bit

// send out our emulated TX byte (time shifted by 1 byte time)
         if (serial_bitcount_tx == serial_databits)
            async_write(coms1, serial_byte_tx);
         else
            {
             printf("serial_w: Break signal sent, serial_bitcount_tx=%d\n", serial_bitcount_tx);             
             async_write_break(coms1);
            }   
        }
    }
}

//==============================================================================
// Serial configuration.
//
// The divider for TX is reduced by 5% and is required by the code section that
// uses it,  a 3% value is not enough.
//
// The calcuations are always based on either a 3.375 or 2 MHz XTAL as these
// were the original speeds.  The correct application software is required
// for each XTAL frequency.  i.e. Old 2 MHz requires old BASIC.
//
//   pass: int cpuclock		XTAL frequency in Hz.
//         int baudtx		TX baud rate
//         int baudrx		RX baud rate
//         int databits         data bits
//         int stopbits         stop bits
// return: void
//==============================================================================
void serial_config (int cpuclock, int baudtx, int baudrx, int databits, int stopbits)
{
 if ((coms1_dev != NULL) && (coms1 != (deschand_t)-1))
    {
     if (cpuclock != 2000000)
        cpuclock = 3375000;

     serial_divval_tx = (int)((float)cpuclock * ((float)(1.0)/baudtx) * (95.0/100.0));
     serial_divval_rx = (int)((float)cpuclock * ((float)(1.0)/baudrx));     

     serial_databits = databits;
     serial_stopbits = stopbits;

     async_configure(coms1, baudtx, baudrx, databits, stopbits, 0);
    } 
}
