//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                                 VDU module                                 *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Emulates the graphics hardware for Standard and Premium models (Alpha+)
//
// - Character ROM
// - Screen RAM
// - PCG RAM
// - Colour RAM
// - Attribute RAM
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.1.0 - 20 October 2007, S.J.Kay
// - Implement the modelx information structure.
//
// v1.4.0 - 30 September 2007, S.J.Kay
// - Added vdu_configure() function to set PCG and VDU values for Premium
//   models.
// - Added Premium PC85 ROM model basic upper/lower half select to vdu_lvdat_w
//   function.
// - Removed the non aplpha+ source build option.
// - Change all error output to stdout.
// - Character ROM loading now uses the memmap_loadrom() function.
// - Added changes to error reporting.
//
// v1.0.0 - 21 June 2007, S.J.Kay
// - Added standard and standard+colour machine options to vdu_colcont_w and
//   vdu_lvdat_w functions.
// - Added vdu_lvdat_r and vdu_lvdat_w port handlers.
// - Implemented Alpha+ (Premium) graphics.
//
// v0.0.0 - 8 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <string.h>
#include <SDL.h>

#include "vdu.h"
#include "crtc.h"
#include "z80.h"
#include "ubee512.h"
#include "memmap.h"
#include "mz80.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
int latchrom;
int attribram;			// LV DAT attribute RAM selected if bit 4=1
int colourram;			// colour RAM selected if bit 4=1 else PCG RAM
int extendram;	        	// LV DAT extended graphics selected if bit 7=1

static int colour_cont;		// colour control port (0x08) value
static int x_colour_cont = -1;
static int x_colourram;

static int videobank;		// LV DAT bank bits 0-3
static int x_videobank;
static int x_attribram;

static int x_extendram;
static int lv_dat;		// LV DAT port (0x1c) value
static int x_lv_dat = -1;

static int vdu_mask;
static int pcg_mask;

static char romimage[512];
char charrom[CHARROMSIZE];

// Alpha+ (Premium) variables, 8K for Screen, Colour and attribute, 32K for PCG
char screen_ram_ap[4 * 0x0800];
char colour_ram_ap[4 * 0x0800];
char attrib_ram_ap[4 * 0x0800];
char pcg_ram_ap[16 * 0x0800];

extern char userhome_romspath[512];
extern int mem_mode;

extern int model;
extern model_t modelx;

//==============================================================================
// VDU initialise
//
//   pass: void
// return: int				0 if success else -1
//==============================================================================
int vdu_init (void)
{
// Alpha+ (Premium) variables, 8K for Screen, Colour and attribute, 32K for PCG
 memset(screen_ram_ap, 0, sizeof(screen_ram_ap));
 memset(colour_ram_ap, 0, sizeof(colour_ram_ap));
 memset(attrib_ram_ap, 0, sizeof(attrib_ram_ap));
 memset(pcg_ram_ap,    0, sizeof(pcg_ram_ap));

 lv_dat=0;		// LV DAT port (0x1c) value
 x_lv_dat=-1;
 videobank = 0;		// LV DATA bank bits 0-3
 attribram = 0;		// LV DATA attribute RAM selected if bit 4=1
 extendram = 0;	        // LV DATA extended graphics selected if bit 7=1

 colour_cont = 0;	// colour control port (0x08) value
 x_colour_cont = -1;
 colourram = 0;		// colour RAM selected if bit 6=1 else PCG RAM

 romimage[0] = 0;
 strcat(romimage, userhome_romspath); 
 strcat(romimage, "charrom.bin"); 

 if (memmap_loadrom(romimage, charrom, CHARROMSIZE) != CHARROMSIZE)
    {
     console_output();
     printf("vdu_init: unable to read %s\n", romimage);
     return -1;
    }
 
 return 0;
}

//==============================================================================
// VDU de-initialise
//
//   pass: void
// return: int				0
//==============================================================================
int vdu_deinit (void)
{
 return 0;
}

//==============================================================================
// VDU reset
//
//   pass: void
// return: int				0
//==============================================================================
int vdu_reset (void)
{
 latchrom = 0;

// Alpha+ (Premium) variables, 8K for Screen, Colour and attribute, 32K for PCG
 memset(screen_ram_ap, 0, sizeof(screen_ram_ap));
 memset(colour_ram_ap, 0, sizeof(colour_ram_ap));
 memset(attrib_ram_ap, 0, sizeof(attrib_ram_ap));
 memset(pcg_ram_ap,    0, sizeof(pcg_ram_ap));

 lv_dat=0;		// LV DAT port (0x1c) value
 x_lv_dat=-1;
 videobank = 0;		// LV DATA bank bits 0-3
 attribram = 0;		// LV DATA attribute RAM selected if bit 4=1
 extendram = 0;	        // LV DATA extended graphics selected if bit 7=1

 colour_cont = 0;	// colour control port (0x08) value
 x_colour_cont = -1;
 colourram = 0;		// colour RAM selected if bit 6=1 else PCG RAM

 return 0;
}

//==============================================================================
// Switch video out.
//
// Must switch out the following video banks from 8000/8800 or F000/F800:
// 
// 1. Copy main memory to attribute/screen RAM (bit 4=1 if attribute) bank
//    (bits 0-3).
// 2. Copy main memory to PCG/Colour RAM bank (bits 0-3) to main memory. The
//    state of port 0x08 will determine if PCG or Colour RAM.
//
// NOTES:
// - if bit 7=0 then banks 0 still must be put back in to place for the above.
// - character ROM is never in main memory so no check is needed for it.
//
//   pass: void
// return: void
//==============================================================================
void vdu_switchvideo_out (void)
{
 int vb;
 int video;

 if (! (mem_mode & BANK_VRAM))		// if video RAM enabled
    {
     if (mem_mode & BANK_VADD)		// if video at 8000-8FFF
        video = 0x8000;
     else
        video = 0xF000;   
  
     if (! x_extendram)	        	// if not extended graphics selected
        vb = 0;
     else    
        vb = x_videobank;
        
     if (x_attribram)             
        memcpy(&attrib_ram_ap[(vb & vdu_mask) * 0x0800], z80mem+video, 0x800);
     else
        memcpy(&screen_ram_ap[(vb & vdu_mask) * 0x0800], z80mem+video, 0x800);
     if (x_colourram)
        memcpy(&colour_ram_ap[(vb & vdu_mask) * 0x0800], z80mem+video+0x0800, 0x800);
     else
        memcpy(&pcg_ram_ap[(vb & pcg_mask) * 0x0800], z80mem+video+0x0800, 0x800);
    }
}

//==============================================================================
// Switch video in.
//
// Must switch in the following video banks to 8000/8800 or F000/F800:
// 
// 1. Copy attribute/screen RAM (bit 4=1 if attribute) bank (bits 0-3) to
//    main memory.
// 2. Copy PCG/Colour RAM bank (bits 0-3) to main memory. The state of port
//    0x08 will determine if PCG or Colour RAM 
//
// NOTES:
// - if bit 7=0 then banks 0 still must be put back in to place for the above.
// - character ROM is never in main memory so no check is needed for it.
//
//   pass: void
// return: void
//==============================================================================
void vdu_switchvideo_in (void)
{
 int vb;
 int video;

 if (! (mem_mode & BANK_VRAM))		// if video RAM enabled
    {
     if (mem_mode & BANK_VADD)		// if video at 8000-8FFF
        video = 0x8000;
     else
        video = 0xF000;   

     if (! extendram)	                // if not extended graphics selected
        vb = 0;
     else    
        vb = videobank;

     if (attribram)             
        memcpy(z80mem+video, &attrib_ram_ap[(vb & vdu_mask) * 0x0800], 0x800);
     else
        memcpy(z80mem+video, &screen_ram_ap[(vb & vdu_mask) * 0x0800], 0x800);
     if (colourram)             
        memcpy(z80mem+video+0x0800, &colour_ram_ap[(vb & vdu_mask) * 0x0800], 0x800);
     else
        memcpy(z80mem+video+0x0800, &pcg_ram_ap[(vb & pcg_mask) * 0x0800], 0x800);

// save the current video state
     x_lv_dat = lv_dat;			// port (0x1c) value
     x_videobank = videobank;		// bank bits 0-3
     x_attribram = attribram;		// attribute RAM select
     x_extendram = extendram; 		// extended graphics select
     x_colour_cont = colour_cont;
     x_colourram = colourram;
    } 
}

//==============================================================================
// Video memory read.
//
// Note: If the 'standard' option is in affect then this code should behave like
// a standard machine. (extendram=0, attribram=0)
//
//   pass: UINT32 addr
//         struct MemoryReadByte *mem_s
// return: UINT8
//==============================================================================
UINT8 vdu_vidmem_r (UINT32 addr, struct MemoryReadByte *mem_s)
{
// if (debuglvl > 0)
//    fprintf(debuglog, "vdu_vidmem_r at 0x%04x\n", (unsigned int)addr);

 int vb;

 if (! extendram)		// if the alpha+ graphics is not enabled
    vb = 0;
 else    
    vb = videobank;

 if (addr & 0x0800)
    {
     if (! colourram)
        {
#if 0
         printf("R_PCG: addr=0x%4x data=0x%2x\n ", addr, pcg_ram_ap[(vb & pcg_mask) * 0x0800 + (addr & 0x07FF)]);     
#endif
         return pcg_ram_ap[(vb & pcg_mask) * 0x0800 + (addr & 0x07FF)];
        }    
     else
        {
#if 0
         printf("R_COL: addr=0x%4x data=0x%2x\n ", addr, colour_ram_ap[(vb & vdu_mask) * 0x0800 + (addr & 0x07FF)]);     
#endif
         return colour_ram_ap[(vb & vdu_mask) * 0x0800 + (addr & 0x07FF)];
        }
    }    
 else
    {
     if (! latchrom)
        {
         if (! attribram)             
            return screen_ram_ap[(vb & vdu_mask) * 0x0800 + (addr & 0x07FF)];
         else   
            return attrib_ram_ap[(vb & vdu_mask) * 0x0800 + (addr & 0x07FF)];
        }
     else       
/* FIXME this will not work if disp_start is set to just before 0x2000, and the
   memory addresses would overlap 0x2000 */
        return charrom[((disp_start & 0x2000) >> 2) + (addr & 0x07FF)];
    }
}

//==============================================================================
// Video memory write.
//
// Note: If the 'standard' option is in affect then this code should behave like
// a standard machine. (extendram=0, attribram=0)
//
//   pass: UINT32 addr
//         UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
void vdu_vidmem_w (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
// if (debuglvl > 0)
//    fprintf(debuglog, "vdu_vidmem_w at 0x%04x = 0x%02x\n", (unsigned int)addr, data);

 int vb;

 if (! extendram)		// if the alpha+ graphics is not enabled
    vb = 0;
 else    
    vb = videobank;

 if (addr & 0x0800)    
    {
     if (! colourram)
        {
#if 0
         printf("W_PCG: addr=0x%4x data=0x%2x\n ", addr, data);
#endif        
         pcg_ram_ap[(vb & pcg_mask) * 0x0800 + (addr & 0x07FF)] = data;
        } 
     else   
        {
#if 0
         printf("W_COL: addr=0x%4x data=0x%2x\n ", addr, data);
#endif         
         colour_ram_ap[(vb & vdu_mask) * 0x0800 + (addr & 0x07FF)] = data; 
        }
     z80mem[addr & 0xFFFF] = data;
     crtc_set_redraw();        
    }    
 else
    {
     if (! latchrom)
        {
         if (! attribram)
            screen_ram_ap[(vb & vdu_mask) * 0x0800 + (addr & 0x07FF)] = data;
         else
            attrib_ram_ap[(vb & vdu_mask) * 0x0800 + (addr & 0x07FF)] = data;
         z80mem[addr & 0xFFFF] = data;
         crtc_redraw_char(addr);
        }
    }
}

//==============================================================================
// Read Port 0x08 - colour control
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			colour_cont
//==============================================================================
UINT16 vdu_colcont_r (UINT16 port, struct z80PortRead *port_s)
{
#if 0 
 printf("vdu_colcont_w: %x\n", colour_cont); 
#endif
 return colour_cont;
}

//==============================================================================
// Write Port 0x08 - colour control
//
// Must switch in either the PCG or Colour RAM banks to 8000/8800 or F000/F800
// depending on the port data.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void vdu_colcont_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (modelx.colour)
    {
     colour_cont = data & B8(11111110);
     colourram = colour_cont & B8(01000000);
#if 0
     printf("vdu_colcont_w: colour_cont=%x colourram=%x\n", colour_cont, colourram);
#endif     
     if (x_colour_cont != colour_cont)
        {
         vdu_switchvideo_out();
         vdu_switchvideo_in();
        } 
    } 
}

//==============================================================================
// Read Port 0x09 - Colour wait off
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UNINT16			0
//==============================================================================
UINT16 vdu_colwait_r (UINT16 port, struct z80PortRead *port_s)
{
#ifdef use_debug_vdu_colwait_r
 printf("Colour wait off (0x09) read"); 
#endif
 return 0;
}

//==============================================================================
// Write Port 0x09 - Colour wait off
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void vdu_colwait_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
#ifdef use_debug_vdu_colwait_w
 printf("Colour wait off (0x09) write"); 
#endif
}

//==============================================================================
// Read Port 0x1C - LV DATA 
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			lv_dat
//==============================================================================
UINT16 vdu_lvdat_r (UINT16 port, struct z80PortRead *port_s)
{
#ifdef use_debug_vdu_lvdat_r
 printf("lvdat (0x1C) read: %x\n", lv_dat);
#endif

 return lv_dat;
}

//==============================================================================
// Write Port 0x1C - LV DATA
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void vdu_lvdat_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
#ifdef use_debug_vdu_lvdat_w
 printf("vdu_lvdat_w: %x\n", lv_dat);
#endif

 if (modelx.standard)
    return;

 lv_dat = data & B8(10111111);		// port (0x1c) value
 videobank = lv_dat & B8(00001111);	// bank bits 0-3
 attribram = lv_dat & B8(00010000);	// attribute RAM select
 extendram = lv_dat & B8(10000000); 	// extended graphics select

 if (x_lv_dat != lv_dat)
    {
     memmap_basicrom(lv_dat & B8(00100000));
     vdu_switchvideo_out();
     vdu_switchvideo_in();
    } 
}

//==============================================================================
// Port 0x0B - Character ROM select
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void vdu_latchrom (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
#ifdef use_debug_vdu_latchrom
 printf("latchrom (0x0B) write: %x\n", latchrom); 
#endif
 latchrom = data & B8(00000001);
}

//==============================================================================
// VDU configure.
//
// Determines the amound of Premium VDU RAM (screen, attribute, colour) and
// amount of PCG RAM.
//
//   pass: int pcg			PCG RAM (16 or 32K)
//         int vdu			VDU RAM (2 or 8 K)
// return: void
//==============================================================================
void vdu_configure (int pcg, int vdu)
{
 pcg_mask = (pcg / 2) - 1; 
 vdu_mask = (vdu / 2) - 1;
}
