//******************************************************************************
//*                                  uBee512                                   *
//*         An emulator for the Microbee Z80 FDD and ROM based models          *
//*                                                                            *
//*                                 z80 module                                 *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Provides an init/deinit and an interface to the MZ80 routines.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v2.1.0 - 28 October 2007, S.J.Kay
// - Implement the modelx information structure.
// - Add clock speed port 0x09 for read.
//
// v2.0.0 - 13 October 2007, S.J.Kay
// - Ports enabled for the Teleterm model and RTC emulation.
//
// v1.4.0 - 26 September 2007, S.J.Kay
// - Boot address (boot_addr) is now dependent on the model being emulated.
// - Added mz80init() to the z80_init() function.  This does not affect 386
//   ASM code builds but is required for 'C' builds of MZ80.  The 'C' build
//   is not currently working on i386 systems and possibly others.
// - Added changes to error reporting.
//
// v1.2.0 - 19 August 2007, S.J.Kay
// - Added a special emulator port to access built in emulator commands.
//
// v1.0.0 - 1 August 2007, S.J.Kay
// - Added alpha+ (Premimum) and other ports with compile time options.
//
// v0.0.0 - 5 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Microbee Z80 FDD and ROM based models. 
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>

#include "z80.h"
#include "ubee512.h"
#include "mz80.h"
#include "crtc.h"
#include "memmap.h"
#include "vdu.h"
#include "fdc.h"
#include "pio.h"
#include "paknet.h"
#include "rtc.h"
#include "keytc.h"
#include "clock.h"
#include "function.h"

UINT16 z80_unhandled_r(UINT16 port, struct z80PortRead *port_s);
void z80_unhandled_w(UINT16 port, UINT8 data, struct z80PortWrite *port_s);

// The Z80 memory
char z80mem[Z80MEMSIZE];

extern model_t modelx;

//==============================================================================
// Define the read ports handlers structure.
//
// Read MZ80 text file documentation for an explanation of this structure.
//
// Any port if used requires an appropriate C module to be compiled in.  The
// first part of the function name indicates which C file the function can
// found in.  The C modules are:
//
// pio.c	Z80 PIO (PC speaker, serial, tape and parallel port).
// rtc.c	Real Time Clock.
// vdu.c	Colour control & waitoff, latch ROM, lvdat (premium graphics).
// paknet.c	NET and PAK selection.
// crtc.c	CRT 6845 emulation.
// soundchip.c	76489AN Sound chip emulation.
// keytc.c	Teleterm keyboard.
// fdc.c	Floppy Disk Controller - WD2793 emulation.
// memmap.c	Memory map handler.
//==============================================================================
static struct z80PortRead z80_ports_r[] =
{
// Z80 PIO (speaker, serial, tape, and parallel port)
 { 0x00, 0x03, pio_r, NULL },
 { 0x10, 0x13, pio_r, NULL },

// RTC
 { 0x04, 0x04, rtc_r, NULL },
 { 0x06, 0x07, rtc_r, NULL },
 { 0x14, 0x14, rtc_r, NULL },
 { 0x16, 0x17, rtc_r, NULL },

// Colour control
 { 0x08, 0x08, vdu_colcont_r, NULL },
#ifndef use_alpha_p
 { 0x18, 0x18, vdu_colcont_r, NULL },
#endif

// CPU clock speed or Viatel video mode
 { 0x09, 0x09, clock_r, NULL },
#ifndef use_alpha_p
 { 0x19, 0x19, clock_r, NULL },
#endif 

// NET selection
 { 0x0A, 0x0A, paknet_nsel_r, NULL },
#ifndef use_alpha_p
 { 0x1A, 0x1A, paknet_nsel_r, NULL },
#endif 

// Video control
 { 0x0C, 0x0C, crtc_status, NULL },
 { 0x0D, 0x0D, crtc_data_r, NULL },
 { 0x0E, 0x0E, crtc_status, NULL },
 { 0x0F, 0x0F, crtc_data_r, NULL },  
#ifndef use_alpha_p
 { 0x1C, 0x1C, crtc_status, NULL },
 { 0x1D, 0x1D, crtc_data_r, NULL },
 { 0x1E, 0x1E, crtc_status, NULL }, 
 { 0x1F, 0x1F, crtc_data_r, NULL },
#else
 { 0x1C, 0x1F, vdu_lvdat_r, NULL },   
#endif

// Soundchip
#ifdef use_soundchip
 { 0x10, 0x13, soundchip_r, NULL },
#endif

// Teleterm keyboard
 { 0x18, 0x1B, keytc_r, NULL },

// Floppy Disk Controller (FDC)
 { 0x40, 0x40, fdc_status, NULL },
 { 0x41, 0x41, fdc_track_r, NULL },
 { 0x42, 0x42, fdc_sect_r, NULL },
 { 0x43, 0x43, fdc_data_r, NULL },
 { 0x44, 0x44, fdc_status, NULL },
 { 0x45, 0x45, fdc_track_r, NULL },
 { 0x46, 0x46, fdc_sect_r, NULL },
 { 0x47, 0x47, fdc_data_r, NULL },
 { 0x48, 0x4B, fdc_ext_r, NULL },

// Special uBee512 built in function commands
 { 0xFF, 0xFF, function_ubee_r, NULL },
 
 { 0x00, 0xFF, z80_unhandled_r, NULL }, 
 
 { -1, -1, NULL, NULL }
};


//==============================================================================
// Define the write ports handlers structure.
//
// Read MZ80 text file documentation for an explanation of this structure.
//
// Any port if used requires an appropriate C module to be compiled in.  The
// first part of the function name indicates which C file the function can
// found in.  The C modules are:
//
// pio.c	Z80 PIO (PC speaker, tape, serial, and parallel port).
// rtc.c	Real Time Clock.
// vdu.c	Colour control & waitoff, latch ROM, lvdat (premium graphics).
// paknet.c	NET and PAK selection.
// crtc.c	CRT 6845 emulation.
// soundchip.c	76489AN Sound chip emulation.
// keytc.c	Teleterm keyboard.
// fdc.c	Floppy Disk Controller - WD2793 emulation.
// memmap.c	Memory map handler.
// clock.c      Clock speed
// function.c   Emulator functions.
//==============================================================================
static struct z80PortWrite z80_ports_w[] =
{
// Z80 PIO (speaker, serial, tape, and parallel port)
 { 0x00, 0x03, pio_w, NULL },
 { 0x10, 0x13, pio_w, NULL },

// RTC
 { 0x04, 0x04, rtc_w, NULL },
 { 0x06, 0x07, rtc_w, NULL },
 { 0x14, 0x14, rtc_w, NULL },
 { 0x16, 0x17, rtc_w, NULL },

// Colour control
 { 0x08, 0x08, vdu_colcont_w, NULL },
#ifndef use_alpha_p
 { 0x18, 0x18, vdu_colcont_w, NULL },
#endif

 // Colour waitoff
 { 0x09, 0x09, vdu_colwait_w, NULL },
#ifndef use_alpha_p
 { 0x19, 0x19, vdu_colwait_w, NULL },
#endif 

// PAK selection
 { 0x0A, 0x0A, paknet_psel_w, NULL },
#ifndef use_alpha_p
 { 0x1A, 0x1A, paknet_psel_w, NULL },
#endif 

// Video control 
 { 0x0B, 0x0B, vdu_latchrom, NULL },
 { 0x0C, 0x0C, crtc_address, NULL },
 { 0x0D, 0x0D, crtc_data_w, NULL },
 { 0x0E, 0x0E, crtc_address, NULL }, 
 { 0x0F, 0x0F, crtc_data_w, NULL },
#ifndef use_alpha_p 
 { 0x1B, 0x1B, vdu_latchrom, NULL },
 { 0x1C, 0x1C, crtc_address, NULL },
 { 0x1D, 0x1D, crtc_data_w, NULL }, 
 { 0x1E, 0x1E, crtc_address, NULL },
 { 0x1F, 0x1F, crtc_data_w, NULL }, 
#else
 { 0x1C, 0x1F, vdu_lvdat_w, NULL },
#endif 

// Soundchip
#ifdef use_soundchip
 { 0x10, 0x13, soundchip_w, NULL },
#endif

// Teleterm keyboard
 { 0x18, 0x1B, keytc_r, NULL },

// Floppy Disk Controller (FDC)
 { 0x40, 0x40, fdc_cmd, NULL },
 { 0x41, 0x41, fdc_track_w, NULL },
 { 0x42, 0x42, fdc_sect_w, NULL },
 { 0x43, 0x43, fdc_data_w, NULL },
 { 0x44, 0x44, fdc_cmd, NULL },
 { 0x45, 0x45, fdc_track_w, NULL },
 { 0x46, 0x46, fdc_sect_w, NULL },
 { 0x47, 0x47, fdc_data_w, NULL },
 { 0x48, 0x4B, fdc_ext_w, NULL },

// Memory map control
 { 0x50, 0x57, memmap_mode, NULL },

// Special uBee512 built in function commands 
 { 0xFF, 0xFF, function_ubee_w, NULL },

 { 0x00, 0xFF, z80_unhandled_w, NULL },

 { -1, -1, NULL, NULL }
};

//==============================================================================
// Z80 Initilization
//
//   pass: void
// return: int				0
//==============================================================================
int z80_init(void)
{
 struct mz80context z80;

 mz80init();
  
 z80.z80Base = (UINT8 *)z80mem;
 z80.z80IoRead = z80_ports_r;
 z80.z80IoWrite = z80_ports_w;

 z80.z80MemRead = NULL;
 z80.z80MemWrite = NULL;

// put interrupt setup stuff here
 mz80SetContext(&z80);

 return 0;
}

//==============================================================================
// Z80 De-Initilization
//
//   pass: void
// return: int				0
//==============================================================================
int z80_deinit(void)
{
 return 0;
}

//==============================================================================
// Z80 Reset
//
//   pass: void
// return: int				0
//==============================================================================
int z80_reset(void)
{
 struct mz80context z80;

 mz80reset();
 mz80GetContext(&z80);
 z80.z80pc = modelx.bootaddr;			// start executing code here
 mz80SetContext(&z80);
 
 return 0;
}

//==============================================================================
// Z80 Get Program Counter (PC)
//
//   pass: void
// return: int				PC
//==============================================================================
int z80_getpc(void)
{
 struct mz80context z80;

 mz80GetContext(&z80);
 return z80.z80pc;
}

//==============================================================================
// Z80 Unhandled read port
//
// Not reported normally as this may be deliberate.  Reporting may affect the
// emulation speed of the program.
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0
//==============================================================================
UINT16 z80_unhandled_r(UINT16 port, struct z80PortRead *port_s)
{
// printf("z80_unhandled_r: Unknown port read: %02X\n", port);
 return 0;
}

//==============================================================================
// Z80 Unhandled write port
//
// Not reported normally as this may be deliberate.  Reporting may affect the
// emulation speed of the program.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================

void z80_unhandled_w(UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
// printf("z80_unhandled_w: Unknown port write: %02X = %02X\n", port, data);
}
